import pytest
import subprocess
import sys

from tackle.models import (
    BaseContext,
    BaseHook,
    Context,
    JinjaHook,
)


def test_base_hook_init():
    """Test instantiating a hook."""
    inp = {
        'hook_type': 'foo',
        # 'foo': 'bar',
        'else': 'foo',
        'merge': True,
        'if': 'stuff == things',
    }

    h = BaseHook(**inp)
    # assert h.hook_type == 'foo'
    assert h.if_ == '{{stuff == things}}'
    assert h.merge
    assert isinstance(h.merge, bool)


@pytest.fixture()
def temporary_uninstall():
    """Fixture to uninstall a package and install it after the test."""

    def f(package):
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "uninstall",
                "--quiet",
                "--disable-pip-version-check",
                "-y",
                package,
            ]
        )

    return f


# https://github.com/robcxyz/tackle-box/issues/47
# def test_imports_import_native_providers(temporary_uninstall):
#     """
#     Validate that when we uninstall requests and import native providers that the
#     imported hook is a LazyImportHook, which when the hook is used will then trigger
#     with a fallback install with the requirements and be imported directly.
#     """
#     temporary_uninstall('requests')
#     pd = ProviderHooks()
#     assert isinstance(pd['http_get'], LazyImportHook)
#     lazy_hook = pd['http_get']
#     pd.import_with_fallback_install(
#         mod_name=lazy_hook.mod_name,
#         path=lazy_hook.hooks_path,
#     )
#     assert issubclass(pd['http_get'], BaseHook)


def test_models_latest():
    """Sanity check for default checkout to latest."""
    c = Context(latest=True)
    assert c.checkout == 'latest'


def test_models_jinja_hook():
    """
    Test that wrapped_exec properly passes references to context vars such as
     public_context between a hook call.
    """
    context = BaseContext()
    context.private_hooks = Context().private_hooks
    c = JinjaHook(context=context, hook=context.private_hooks['set'])
    c.context.public_context = {}
    c.wrapped_exec(path="foo", value=1)
    assert c.context.public_context == {"foo": 1}



from tackle import tackle


def test_render_globals_base(change_curdir_fixtures):
    """
    Verify that when we have a variable in the context, for instance `namespace`, that
     we are able to render it from the context.
    """
    output = tackle('globals.yaml')
    for i in output['globals_raw']:
        assert i == 'stuff'


def test_render_hooks_with_args(change_curdir_fixtures):
    """Verify that we can call hooks without args."""
    output = tackle('hooks-args.yaml')
    assert output['get_default'] == 'things'
    assert output['get_sep'] == 'things'
    assert 'tmp' in output['nested']
    assert output['nested_test']['and'] == 'things'



import os

from tackle.models import BaseHook, Field


class ListdirHook(BaseHook):
    """Hook for `listdir`. Lists the contents of a directory."""

    hook_type: str = 'listdir'
    path: str = Field(None, description="Path to directory to list contents.")
    sort: bool = Field(False, description="Boolean to sort the output")
    ignore_hidden_files: bool = Field(
        None, description="Boolean to ignore hidden files"
    )

    args: list = ['path']
    _docs_order = 2

    def exec(self) -> list:
        files = os.listdir(os.path.expanduser(self.path))
        if self.sort:
            files.sort()
        if self.ignore_hidden_files:
            return [f for f in files if not f.startswith('.')]
        else:
            return files



from pydantic import SecretStr
from typing import Union

import tackle as tkl
from tackle.models import BaseHook, Field


class TackleHook(BaseHook):
    """Hook for calling external tackle providers."""

    hook_type: str = 'tackle'

    # fmt: off
    input_string: str = Field(
        None,
        description="The input can be one of repo, file path, directory with "
                    "tackle.yaml, zip file, or if left blank parent tackle file."
    )
    checkout: str = Field(
        None,
        description="The branch or version to checkout for repo type inputs_strings."
    )
    latest: bool = Field(
        False,
        description="For remote providers, use the latest commit."
    )
    context_file: str = Field(None, description="The file to run inside a repo input.")
    extra_context: dict = Field(
        None,
        description="Any additional context to use when calling the hook. Like existing context.")
    context: dict = Field(
        None,
        description="A context to use that overrides the current context."
    )
    password: SecretStr = Field(None, description="A password to use for repo inputs.")
    directory: str = Field(
        None,
        description="The directory to run inside for repo inputs."
    )
    find_in_parent: bool = Field(
        False,
        description="Search for target in parent directory. Only relevant for local "
                    "targets.")

    override: dict = Field({}, description="A dictionary of keys to override.")

    additional_args: Union[list, str] = Field(
        None,
        description="Arguments to pass on either directly as a string or as a list of "
                    "strings."
    )
    # fmt: on

    args: list = ['input_string', 'additional_args']
    kwargs: str = 'override'
    _docs_order = 0

    def exec(self) -> dict:
        existing_context = {}

        if self.override != {}:
            # This is a hack as we need extra vars to be mapped to both the
            # existing_context (allowing rendering of unknown vars) and the overrides
            # (allowing for calling of declarative hooks with args / kwargs)
            existing_context.update(self.override)

        if self.context:
            existing_context.update(self.context)

        if self.existing_context:
            existing_context.update(self.existing_context)

        if self.temporary_context:
            existing_context.update(self.temporary_context)

        if self.private_context:
            existing_context.update(self.private_context)

        if self.public_context:
            existing_context.update(self.public_context)

        if self.extra_context:
            existing_context.update(self.extra_context)

        if isinstance(self.additional_args, str):
            self.additional_args = [self.additional_args]

        output_context = tkl.main.tackle(
            self.input_string,
            checkout=self.checkout,
            latest=self.latest,
            password=self.password,
            directory=self.directory,
            calling_directory=self.calling_directory,
            calling_file=self.calling_file,
            # Evaluated
            existing_context=existing_context,
            # Implicit
            public_hooks=self.public_hooks,
            private_hooks=self.private_hooks,
            no_input=self.no_input,
            global_kwargs=self.override,
            global_args=self.additional_args,
            find_in_parent=self.find_in_parent,
            verbose=self.verbose,
        )

        return output_context



import re

from tackle.models import BaseHook, Field


class StartsWithHook(BaseHook):
    """
    Hook checking if a string starts with another string.
    [Docs](https://www.tutorialspoint.com/python/string_startswith.htm)
    """

    hook_type: str = 'starts_with'

    input: str = Field(..., description="A string to check")
    match: str = Field(..., description="The chars to match")

    beg: int = Field(
        None,
        description="This is the optional parameter to set start "
        "index of the matching boundary.",
    )
    end: int = Field(
        None,
        description="This is the optional parameter to end start "
        "index of the matching boundary.",
    )

    args: list = ['input', 'match']
    _docs_order = 8

    def exec(self):
        return self.input.startswith(self.match)


class EndsWithHook(BaseHook):
    """
    Hook checking if a string ends with another string.
    [Docs](https://www.tutorialspoint.com/python/string_startswith.htm)
    """

    hook_type: str = 'ends_with'

    input: str = Field(..., description="A string to check")
    match: str = Field(..., description="The chars to match")

    beg: int = Field(
        None,
        description="This is the optional parameter to set start "
        "index of the matching boundary.",
    )
    end: int = Field(
        None,
        description="This is the optional parameter to end start "
        "index of the matching boundary.",
    )

    args: list = ['input', 'match']
    _docs_order = 8

    def exec(self):
        return self.input.endswith(self.match)


class RegexMatchHook(BaseHook):
    """
    Hook running a regex against a string.
    [Docs](https://docs.python.org/3/library/re.html#re.match)
    """

    hook_type: str = 'starts_with'

    pattern: str = Field(..., description="A regex pattern to check.")
    string: str = Field(..., description="A string to check.")

    args: list = ['pattern', 'string']

    def exec(self):
        return re.match(self.pattern, self.string)



from tackle import BaseHook, Field
import paramiko


class SshInteractiveQuery(BaseHook):
    """Interactive shell session hook."""

    hook_type = "ssh_interactive"

    # fmt: off
    host: str = Field(None, description="Host address.")
    username: str = Field(..., description="User name used to authenticate.")
    port: int = Field(22, description="Connection port number.")
    password: str = Field(None, description="Optional password used to authenticate.")
    key_filename: str = Field(None, description="Private key file path.")

    # fmt: on

    def interactive(self):
        pass

    def exec(self):
        try:
            import interactive
        except ImportError:
            from . import interactive

        client = paramiko.SSHClient()

        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        client.connect(
            self.host,
            username=self.username,
            key_filename=self.key_filename
        )
        channel = client.get_transport().open_session()
        channel.get_pty()
        channel.invoke_shell()

        interactive.interactive_shell(channel)
        client.close()



import pytest
from ruamel.yaml import YAML

import shutil
import os

from tackle.main import tackle
from tackle.providers.generate.hooks.exceptions import (
    UndefinedVariableInTemplate,
    GenerateHookTemplateNotFound,
)

FIXTURES = [
    "file.yaml",
    "plain-src.yaml",
    "plain-src-block.yaml",
    "plain-src-path.yaml",
    "render-file.yaml",
    "render-file-additional-context.yaml",
    "render-dir-file.yaml",
    "render-dir-file-base.yaml",
    # "tackle-provider-remote.yaml",
]


@pytest.mark.parametrize("fixture", FIXTURES)
def test_provider_system_hook_generate_fixtures(change_dir, fixture):
    output = tackle(fixture, no_input=True)
    assert not output['init']

    if isinstance(output['after'], list):
        for i in output['after']:
            assert i
    else:
        assert output['after']


ERRORS = [
    ("missing-file.yaml", GenerateHookTemplateNotFound),
    ("unknown-variable.yaml", UndefinedVariableInTemplate),
]


@pytest.mark.parametrize("fixture,error", ERRORS)
def test_provider_system_hook_generate_error(change_dir, fixture, error):
    with pytest.raises(error):
        tackle(fixture)


def test_hook_generate_copy_without_render(change_dir):
    tackle("copy-without-render.yaml")
    yaml = YAML()
    with open(os.path.join('output', '.hidden.yaml')) as f:
        hidden = yaml.load(f)

    with open('output/no-render/dir/foo.yaml') as f:
        nested_glob = yaml.load(f)

    assert hidden['stuff'] == '{{stuff}}'
    assert nested_glob['stuff'] == '{{stuff}}'
    shutil.rmtree('output')


def test_hook_generate_looped(change_dir):
    output = tackle("looped.yaml")

    assert len(output['networks']) == 2
    assert output['networks'][0]['things'] == 'foo'
    shutil.rmtree('output')


def test_hook_generate_skip_overwrite_files(change_dir):
    """Validate that we can skip rendering a file with `skip_overwrite_files`."""
    output = tackle("skip-overwrite-files.yaml")

    assert output['verify']['stuff'] == 'foo'


def test_hook_generate_skip_if_file_exists(change_dir):
    """Validate that we can skip rendering a file with `skip_if_file_exists`."""
    output = tackle("skip-overwrite-files.yaml")

    assert output['verify']['stuff'] == 'foo'



from tackle import tackle


def test_provider_logic_type(change_dir):
    """Check type works."""
    output = tackle('type.yaml')
    assert output['str_type'] == 'str'
    assert output['list_type'] == 'list'
    assert output['a_dict_type'] == 'dict'
    assert output['dict_list_type'] == 'dict'
    assert output['list_dict_type'] == 'list'
    assert output['with_args']['type'] == 'list'
    assert output['default']['type'] == 'list'


def test_hook_casting(change_dir):
    """Check casting works."""
    output = tackle('castings.yaml')
    # Assertions in file
    assert output



import os
import platform
import csv
from typing import TYPE_CHECKING
from xdg import (
    xdg_cache_home,
    xdg_config_dirs,
    xdg_config_home,
    xdg_data_dirs,
    xdg_data_home,
    xdg_runtime_dir,
    xdg_state_home,
)

from tackle.settings import settings

if TYPE_CHECKING:
    from tackle.models import Context


def get_linux_distribution():
    """Return the equivalent of lsb_release -a."""
    if platform.system() == 'Linux':
        RELEASE_DATA = {}
        with open("/etc/os-release") as f:
            reader = csv.reader(f, delimiter="=")
            for row in reader:
                if row:
                    RELEASE_DATA[row[0]] = row[1]
        if RELEASE_DATA["ID"] in ["debian", "raspbian"]:
            with open("/etc/debian_version") as f:
                DEBIAN_VERSION = f.readline().strip()
            major_version = DEBIAN_VERSION.split(".")[0]
            version_split = RELEASE_DATA["VERSION"].split(" ", maxsplit=1)
            if version_split[0] == major_version:
                # Just major version shown, replace it with the full version
                RELEASE_DATA["VERSION"] = " ".join([DEBIAN_VERSION] + version_split[1:])
        return "{} {}".format(RELEASE_DATA["NAME"], RELEASE_DATA["VERSION"])
    else:
        return None


def _cwd():
    return os.getcwd()


def _home_dir():
    return os.path.expanduser('~')


def _system():
    return platform.system()


def _platform():
    return platform.platform()


def _release():
    return platform.release()


def _version():
    return platform.version()


def _processor():
    return platform.processor()


def _architecture():
    return platform.architecture()


def _tackle_dir():
    return settings.tackle_dir


def _provider_dir():
    return settings.provider_dir


def _calling_directory(context: 'Context'):
    return context.calling_directory


def _calling_file(context: 'Context'):
    return context.calling_file


def _current_file(context: 'Context'):
    return context.input_file


def _current_directory(context: 'Context'):
    return context.input_dir


def _this(context: 'Context'):
    return dict(context.public_context)


def _public_context(context: 'Context'):
    return context.public_context


def _private_context(context: 'Context'):
    return context.private_context


def _temporary_context(context: 'Context'):
    return context.temporary_context


def _existing_context(context: 'Context'):
    return context.existing_context


def _key_path(context: 'Context'):
    return context.key_path


def _key_path_block(context: 'Context'):
    return context.key_path_block


def _xdg_cache_home():
    return xdg_cache_home()


def _xdg_config_dirs():
    return xdg_config_dirs()


def _xdg_config_home():
    return xdg_config_home()


def _xdg_data_dirs():
    return xdg_data_dirs()


def _xdg_data_home():
    return xdg_data_home()


def _xdg_runtime_dir():
    return xdg_runtime_dir()


def _xdg_state_home():
    return xdg_state_home()


special_variables = {
    'cwd': _cwd,
    'home_dir': _home_dir,
    'system': _system,
    'platform': _platform,
    'version': _version,
    'processor': _processor,
    'architecture': _architecture,
    'lsb_release': get_linux_distribution,
    'tackle_dir': _tackle_dir,
    'provider_dir': _provider_dir,
    'calling_directory': _calling_directory,
    'calling_file': _calling_file,
    'current_file': _current_file,
    'current_directory': _current_directory,
    'this': _this,
    'public_context': _public_context,
    'private_context': _private_context,
    'existing_context': _existing_context,
    'temporary_context': _temporary_context,
    'key_path': _key_path,
    'key_path_block': _key_path_block,
    'xdg_cache_home': _xdg_cache_home,
    'xdg_config_dirs': _xdg_config_dirs,
    'xdg_config_home': _xdg_config_home,
    'xdg_data_dirs': _xdg_data_dirs,
    'xdg_data_home': _xdg_data_home,
    'xdg_runtime_dir': _xdg_runtime_dir,
    'xdg_state_home': _xdg_state_home,
}



import os
from ruamel.yaml import YAML

from tackle.main import tackle
import pytest


@pytest.fixture()
def clean_things():
    """Remove all things.py."""
    if os.path.exists('things.py'):
        os.remove('things.py')
    yield
    if os.path.exists('things.py'):
        os.remove('things.py')


def test_provider_system_hook_jinja(change_dir, clean_things):
    context = tackle()
    assert context['foo'] == 'bar'
    yaml = YAML()
    with open('things.py') as f:
        output = yaml.load(f)
    assert output == "x = 'bar'"


def test_provider_system_hook_jinja_existing_context(change_dir, clean_things):
    context = tackle('existing-context.yaml')
    assert context['foo'] == 'bar'
    yaml = YAML()
    with open('things.py') as f:
        output = yaml.load(f)
    assert output == "x = 'bar'"



import random
import string
from pydantic import validator

from tackle.models import BaseHook, Field
from tackle.exceptions import HookCallException


class RandomStringHook(BaseHook):
    """Hook for `random_string`. Lists the contents of a directory."""

    # TODO: Update this interface?  Rm string_case replace with --lower/upper flag?

    hook_type: str = 'random_string'
    length: int = Field(8, description="Length of the random string.")
    case: str = Field('lower', description="Case of output, one of `upper` or `lower`")
    upper: bool = Field(False, description="Flag for upper case. Overrides `case`.")

    args: list = ['length', 'case']
    _docs_order = 6

    @validator('case')
    def validate_string_case(cls, v):
        if v not in ['upper', 'lower']:
            raise HookCallException(
                "Field `string_case` must be one of 'upper' or 'lower'."
            )
        return v

    def exec(self) -> str:
        if self.upper:
            self.case = 'upper'

        if self.case == 'upper':
            return ''.join(
                random.choices(string.ascii_uppercase + string.digits, k=self.length)
            )
        elif self.case == 'lower':
            return ''.join(
                random.choices(string.ascii_lowercase + string.digits, k=self.length)
            )


class RandomHexHook(BaseHook):
    """Hook  for `random_hex`. Lists the contents of a directory."""

    hook_type: str = 'random_hex'
    length: int = Field(8, description="Number for number of digits - default 8")

    args: list = ['length']
    _docs_order = 7

    def exec(self) -> str:
        return ''.join(['%0', str(self.length), 'x']) % random.randrange(
            16 ** self.length
        )



"""Match hook."""
from typing import Union
import re

from tackle.models import BaseHook, Context, Field
from tackle.parser import walk_sync
from tackle.render import render_string
from tackle.exceptions import HookCallException


class MatchHook(BaseHook):
    """
    Hook for match / case statements. Takes a dict where the keys are matched to a
     value. If the case value has an arrow in it (ie key->: ... ) the arrow is stripped
     away. All matched values are ran as hooks.
    """

    hook_type: str = 'match'
    value: str = Field(
        ..., render_by_default=True, description="The value to match against."
    )
    case: dict = Field(
        ...,
        description="A dictionary where the keys are cases to be matched. Runs hooks "
        "if present.",
    )

    args: list = ['value']

    _render_exclude = {'case'}
    _docs_order = 3

    def block_macro(self, key, val) -> dict:
        """Take input and create a block hook to parse."""
        output = {key[-2:]: 'block', 'merge': True}
        aliases = [v.alias for _, v in BaseHook.__fields__.items()] + ['->', '_>']
        for k, v in val.items():
            if k not in aliases:
                # Set the keys under the `items` key per the block hook's input
                output.update({'items': {k: v}})
            else:
                output.update({k: v})
        if self.verbose:
            print(
                "You are likely going to hit a bug."
                "https://github.com/robcxyz/tackle-box/issues/67"
            )
        return {key[:-2]: output}
        # return output

    def exec(self) -> Union[dict, list]:
        # Condition catches everything except expanded hook calls and blocks (ie key->)
        for k, v in self.case.items():
            if re.match(k, self.value):
                # Dicts that are not expanded hooks themselves
                if isinstance(v, (dict, list)) and not ('->' in v or '_>' in v):
                    return self.run_key(v)
                elif isinstance(v, dict):
                    # return self.run_key({k: {**v, **{'merge': True}}})
                    return self.run_key({k: {**v, **{'merge': True}}})
                elif isinstance(v, str):
                    return render_string(self, v)
                return v

            elif re.match(k[:-2], self.value) and k[-2:] in ('->', '_>'):
                # Return the value indexed without arrow
                if isinstance(v, str):
                    return self.run_key({k[:-2]: {k[-2:]: v + ' --merge'}})
                elif isinstance(v, dict):
                    return self.run_key(self.block_macro(k, v))
                else:
                    raise HookCallException(
                        f"Matched value must be of type string or dict, not {v}.",
                        context=self,
                    ) from None

        raise Exception(f"Value `{self.value}` not found in {self.case.keys()}")

    def run_key(self, value):
        self.skip_output: bool = True

        tmp_context = Context(
            input_context=value,
            key_path=self.key_path.copy(),
            key_path_block=self.key_path.copy(),
            # provider_hooks=self.provider_hooks,
            public_hooks=self.public_hooks,
            private_hooks=self.private_hooks,
            public_context=self.public_context,
            private_context=self.private_context,
            temporary_context=self.temporary_context,
            existing_context=self.existing_context,
            no_input=self.no_input,
            calling_directory=self.calling_directory,
            calling_file=self.calling_file,
            verbose=self.verbose,
        )
        walk_sync(context=tmp_context, element=value.copy())

        return self.public_context



import os
from typing import Union

from tackle.models import Context
from tackle.parser import update_source
from tackle.utils.paths import find_nearest_tackle_file
from tackle.utils.files import read_config_file
from tackle import exceptions


def get_global_kwargs(kwargs):
    """Check for unknown kwargs and return so they can be consumed later."""
    global_kwargs = {}
    for k, v in kwargs.items():
        if k not in Context.__fields__:
            global_kwargs.update({k: v})
    if global_kwargs == {}:
        return None
    return global_kwargs


def tackle(
    *args,
    **kwargs,
) -> Union[dict, list]:
    """
    Run Tackle Box just as if using it from the command line.

    :param checkout: The branch, tag or commit ID to checkout after clone.
    :param no_input: Prompt the user at command line for manual configuration?
    :param context_file: The file to load to set the context, ie list of prompts.
        Defaults to tackle.yaml then cookiecutter.json.
    :param existing_context: A string pointing to a json/yaml or dictionary to use as
        additional context when rendering items.
    :param password: The password to use when extracting the repository.
    :param directory: Relative path to a cookiecutter template in a repository.

    :return: Dictionary of the output
    """
    if args:
        # Remove first arg and consider all other args global to be ingested later
        kwargs['input_string'] = args[0]
        if len(args) != 1:
            kwargs['global_args'] = []
            for i in range(1, len(args)):
                kwargs['global_args'].append(args[i])

    # Handle empty calls which fallback to finding the closest tackle file
    # in the parent directory
    if 'input_string' not in kwargs or kwargs['input_string'] is None:
        kwargs['input_string'] = find_nearest_tackle_file()

    # Handle the exception if no tackle file is found in parent directory
    if kwargs['input_string'] is None:
        raise exceptions.NoInputOrParentTackleException(
            "No input or tackle file has been given."
        ) from None

    # Initialize context
    context = Context(**kwargs)
    if 'global_kwargs' not in kwargs:
        # When tackle is called from the cli it already has the kwargs.  This is a hack
        # to get this function to accept dicts the same as they would be for use as a
        # package - ie tackle('input-file.yaml', **some_override_dict)
        context.global_kwargs = get_global_kwargs(kwargs)

    if 'override' in kwargs and kwargs['override'] is not None:
        overrides = kwargs.pop('override')
        if isinstance(overrides, str):
            if os.path.exists(overrides):
                override_dict = read_config_file(overrides)
                if override_dict is not None:
                    context.global_kwargs.update(override_dict)
            else:
                raise exceptions.UnknownInputArgumentException(
                    f"The `override` input={overrides}, when given as a string must "
                    f"be a path to an file. Exiting.",
                    context=context,
                )
        elif isinstance(overrides, dict):
            context.global_kwargs.update(overrides)

    # Main loop
    output = update_source(context)
    if output is None:
        return context.public_context
    return output



from tackle.main import tackle


def test_provider_system_hook_var(change_dir):
    output = tackle('var.yaml')
    assert output['short'] == output['a_list']
    assert output['embed'] == "{{a_var_that_doesnt_exist}}"
    assert output['list_of_lists'] == [[1, 2], [3, 4]]
    assert output['a_map_rendered']['stuff'] == 'things'



# from tackle.main import tackle
#
# TODO: https://github.com/robcxyz/tackle-box/issues/46
# def test_provider_pyinquirer_confirm_hook(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.confirm.prompt', return_value={"tmp": True}
#     )
#     o = tackle()
#     assert o['this']



"""Tests dict input objects for `tackle.providers.system.hooks.var` module."""
from tackle.main import tackle


def test_provider_strings_split(change_dir):
    output = tackle('split.yaml')

    assert output['compact'] == output['expanded']
    assert output['compact'][0] == 'stuff'


def test_provider_strings_join(change_dir):
    output = tackle('join.yaml')

    assert output['a_str'] == ['things', 'stuffs']
    assert output['join_a_str'] == 'things.stuffs'
    # See test_render_hook_call_multiple for other test on this next assertion
    assert output['apath'] == "stuff.txt"



import os
import configparser
from typing import Union

from tackle.models import BaseHook, Field


class IniHook(BaseHook):
    """
    Hook for reading and writing ini files. Hook reads from `path` if no `data` field is
     provided, otherwise it writes the `data` to `path`.
    """

    hook_type: str = 'ini'
    path: str = Field(..., description="The file path to put read or write to.")
    data: Union[dict, list, str] = Field(
        None,
        description="Map or renderable string to a map key to write. Must have three "
        "levels for the section, key, and value - ie "
        "{section:{key1:value1, key2:value2}}.",
        render_by_default=True,
    )
    allow_no_value: bool = Field(True, description="Whether to allow a no values.")
    args: list = ['path', 'data']

    def exec(self) -> Union[dict, str, list]:
        self.path = os.path.abspath(os.path.expanduser(os.path.expandvars(self.path)))
        # Make path if it does not exist
        if not os.path.exists(os.path.dirname(self.path)) and self.data:
            os.makedirs(os.path.dirname(self.path))

        config = configparser.ConfigParser(allow_no_value=self.allow_no_value)

        if self.data:
            for section, items in self.data.items():
                config.add_section(section)
                for k, v in items.items():
                    if v is None and self.allow_no_value:
                        config.set(section, k)
                    else:
                        config.set(section, k, str(v))
            with open(self.path, 'w') as f:
                config.write(f)

            return self.path

        else:
            output = {}
            config.read(self.path)
            for section in config.sections():
                output[section] = {}
                for option in config.options(section):
                    output[section][option] = config.get(section, option)

            return output



from typing import Any
from pydantic import Field

from tackle.models import BaseHook
from tackle.render import render_variable


class VarHook(BaseHook):
    """
    Hook for rendering a variable based on an input. Only useful for rendering as
     otherwise you wouldn't need this hook at all. Does recursion when the value being
     rendered is still renderable - ie a template within a template.
    """

    hook_type: str = 'var'
    input: Any = Field(..., description="Any variable input.", render_by_default=True)
    no_recursion: bool = Field(
        False, description="Don't recursively render embedded templates."
    )

    args: list = ['input']

    def _render_var(self, input) -> Any:
        output = render_variable(self, input)
        if output == input:
            return output
        else:
            return self._render_var(output)

    def exec(self) -> Any:
        if self.no_recursion:
            # Input is already rendered so only need to pass back
            return self.input
        else:
            return self._render_var(self.input)



from kubernetes import client, config, utils

from tackle.models import BaseHook, Field


class K8sApplyHook(BaseHook):
    """Hook for getting current kubeconfig context."""

    hook_type: str = 'k8s_apply'

    yaml_dir: str = Field(...)

    def exec(self):
        config.load_kube_config()
        k8s_client = client.ApiClient()
        utils.create_from_directory(k8s_client, self.yaml_dir, verbose=True)



from tackle.main import tackle


def test_provider_prompt_checkbox_map_normal_no_input(change_dir):
    output = tackle('map_normal_checked.yaml', no_input=True)
    assert output['selection'] == ['stuff', 'things']


def test_provider_prompt_checkbox_map_no_input(change_dir):
    output = tackle('map_checked.yaml', no_input=True)
    assert output['selection'] == ['stuff', 'things']


def test_provider_prompt_checkbox_list_no_input(change_dir):
    output = tackle('list_checked.yaml', no_input=True)
    assert output['selection'] == ['stuff', 'things']


# # TODO: https://github.com/robcxyz/tackle-box/issues/46
# def test_provider_prompt_checkbox_list(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.checkbox.prompt',
#         return_value={"tmp": ["things"]},
#         autospec=True,
#     )
#
#     output = tackle('list.yaml')
#     assert output['selection'] == ['things']
#
#
# def test_provider_prompt_checkbox_list_index(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.checkbox.prompt',
#         return_value={"tmp": ["things"]},
#     )
#     output = tackle('list_index.yaml')
#     assert output['selection'] == [1]
#
#
# def test_provider_prompt_checkbox_map(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.checkbox.prompt',
#         return_value={"tmp": ["I do things"]},
#     )
#     output = tackle('map.yaml')
#     assert output['selection'] == ['things']
#
#
# def test_provider_prompt_checkbox_map_normal(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.checkbox.prompt',
#         return_value={"tmp": ["things"]},
#     )
#     output = tackle('map_normal.yaml')
#     assert output['selection'] == ['things']
#
#
# def test_provider_prompt_checkbox_map_index(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.checkbox.prompt',
#         return_value={"tmp": ["I do things"]},
#     )
#     output = tackle('map_index.yaml')
#     assert output['selection'] == [1]



from typing import List

from tackle.models import BaseHook, Field


class SplitHook(BaseHook):
    """Hook for splitting a string into as list based on a separator."""

    hook_type: str = 'split'

    input: str = Field(..., description="A string to split into a list")
    separator: str = Field("/", description="String separator")

    args: list = ['input', 'separator']
    _docs_order = 2

    def exec(self):
        return self.input.split(self.separator)


class JoinHook(BaseHook):
    """Join a list of strings with a separator."""

    hook_type: str = 'join'

    input: List[str] = Field(
        ..., description="A list of strings to join.", render_by_default=True
    )
    separator: str = Field('', description="String separator.")

    args: list = ['input', 'separator']
    _docs_order = 3

    def exec(self):
        return self.separator.join(self.input)



from rich.console import Console
from rich.table import Table
from typing import List, Any

from tackle.models import BaseHook, Field


class TableHook(BaseHook):
    """Hook for creating tables with rich - github.com/willmcgugan/rich."""

    hook_type: str = 'table'

    column_names: List = Field([], description="List of column names")
    sort: bool = Field(False, description="Boolean to sort contents or contents_split")
    contents: Any = Field(None, description="List of lists to put into columns / rows")
    contents_split: List = Field(
        None,
        description="List of strings to separate into columns based on `separator`",
    )
    separator: str = Field(
        None, description="A string to separate the strings in the contents"
    )

    def exec(self):
        if self.sort:
            if self.contents:
                self.contents.sort()
            if self.contents_split:
                self.contents_split.sort()

        console = Console()

        if len(self.column_names) > 0:
            table = Table(show_header=True, header_style="bold red")
            for c in self.column_names:
                table.add_column(c)
        else:
            table = Table()

        if self.contents_split:
            for i in self.contents_split:
                table.add_row(
                    *i.split(self.separator)[0 : len(self.column_names)]  # noqa
                )

        if self.contents:
            for i in self.contents:
                table.add_row(*[str(j) for j in i])

        console.print(table)



"""Tests dict input objects for `tackle.providers.system.hooks.variable` module."""
from tackle.main import tackle
import pytest
import os
import shutil


@pytest.fixture()
def clean_outputs():
    """Clean the outputs."""
    yield
    files = ["things.zip", "things"]
    for f in files:
        if os.path.isfile(f):
            os.remove(f)
        if os.path.isdir(f):
            shutil.rmtree(f)


def test_provider_system_zip_dir(change_dir, clean_outputs):
    output = tackle('zip-dir.yaml')

    assert os.path.isfile(output['dir'])


def test_provider_system_zip_file(change_dir, clean_outputs):
    output = tackle('zip-file.yaml')

    assert os.path.isfile(output['file'])


def test_provider_system_zip_unzip(change_dir, clean_outputs):
    output = tackle('unzip.yaml')

    assert os.path.isdir(output['dir'])



from tackle.main import tackle

import pytest
import shutil


@pytest.fixture()
def clean_outputs():
    """Remove the outputs dir."""
    yield
    shutil.rmtree('output')


# def test_provider_system_hook_tackle(change_dir):
# #     # TODO Build example repo
#     context = tackle('tackle.yaml', no_input=True)
#     assert context


def test_provider_tackle_local(change_dir):
    output = tackle('local.yaml', no_input=True)
    assert output['shell']['stuff'] == 'bing'


def test_provider_tackle_local_no_context(change_dir):
    output = tackle('local-no-context.yaml', no_input=True)
    assert output['shell']['foo'] == 'bar'


def test_provider_tackle_local_prior_context(change_dir):
    output = tackle('local-prior-context.yaml', no_input=True)
    assert output['shell']['foo'] == 'bar'


def test_provider_tackle_block_tackle(change_dir):
    output = tackle('block-tackle.yaml', no_input=True)
    assert 'things' in output


def test_provider_tackle_remote(change_dir, clean_outputs):
    output = tackle('remote.yaml', no_input=True)
    assert output['project_slug'] == 'output'


def test_provider_tackle_kwargs_default(change_dir):
    output = tackle('kwargs-default.yaml', no_input=True)
    assert output['default_kwargs']['stuff'] == 'bing'


def test_provider_tackle_kwargs_default_hook(change_dir):
    """Check that we can run a default hook from a tackle hook."""
    output = tackle('kwargs-default-hook.yaml', no_input=True)
    assert output['compact']['v'] == 'bing'
    assert output['expanded']['v'] == 'bing'


def test_provider_tackle_kwargs_default_hook_args(change_dir):
    """Check that we can run a default hook from a tackle hook with args."""
    output = tackle('kwargs-default-hook-arg.yaml', no_input=True)
    assert output['compact']['v'] == 'bing'



from jinja2.exceptions import UndefinedError
from jinja2 import FileSystemLoader

from typing import Union

from tackle import BaseHook, Field
from tackle.providers.generate.hooks.exceptions import UndefinedVariableInTemplate


class JinjaHook(BaseHook):
    """
    Hook for jinja templates. If given an `output`, the rendered contents are output to
     a file, otherwise the rendered contents are output as a string.
    """

    hook_type: str = 'jinja'
    # fmt: off
    template: str = Field(...,
                          description="Path to the template to render relative to `file_system_loader`.")
    output: str = Field(None, description="Path to the output the template.")
    extra_context: dict = Field(None,
                                description="Extra context update the global context to render with.")
    render_context: dict = Field(
        None, description="A render context that invalidates the default context."
    )
    additional_context: dict = Field(
        None, description="A map to use as additional context when rendering."
    )
    file_system_loader: Union[str, list] = Field(
        '.',
        description="List of paths or string path to directory with templates to load "
                    "from. [Docs](https://jinja.palletsprojects.com/en/3.0.x/api/#jinja2.FileSystemLoader).")
    # fmt: on

    args: list = ['template', 'output']

    def _init_context(self):
        # Update the render_context that will be used
        if self.render_context is not None:
            return

        # fmt: off
        existing_context = self.existing_context if self.existing_context is not None else {}
        temporary_context = self.temporary_context if self.temporary_context is not None else {}
        private_context = self.private_context if self.private_context is not None else {}
        public_context = self.public_context if self.public_context is not None else {}
        # fmt: on

        self.render_context = {
            **existing_context,
            **temporary_context,
            **private_context,
            **public_context,
        }

        if self.extra_context is not None:
            if isinstance(self.extra_context, list):
                for i in self.extra_context:
                    self.render_context.update(i)
            else:
                self.render_context.update(self.extra_context)

    def exec(self) -> str:
        self._init_context()
        self.env_.loader = FileSystemLoader(self.file_system_loader)
        template = self.env_.get_template(self.template)

        try:
            output_from_parsed_template = template.render(**self.render_context)
        except UndefinedError as err:
            msg = f"The `jinja` hook failed to render -> {err}"
            raise UndefinedVariableInTemplate(msg, hook=self) from None

        if self.output is not None:
            with open(self.output, 'w') as fh:
                fh.write(output_from_parsed_template)
            return self.output
        else:
            return output_from_parsed_template



import os
from typing import Union

from tackle.models import BaseHook, Field


class FlattenPathHook(BaseHook):
    """Hook for flattening a dictionary into a list of paths."""

    hook_type: str = 'flatten_paths'
    paths: Union[dict, list] = Field(
        ...,
        description="A list or map representing paths to files that should be flattened",
    )
    base_path: str = Field('', description="A base path to prefix in the output.")

    args: list = ['paths']

    def flatten_repo_tree(self, output: list, d, parent_key=''):
        """Flatten a dict to so that keys become nodes in a path."""
        if isinstance(d, dict):
            for k, v in d.items():
                new_key = os.path.join(parent_key, k)
                if isinstance(v, str):
                    output.append(os.path.join(new_key, v))
                else:
                    self.flatten_repo_tree(output, v, new_key)

        elif isinstance(d, list):
            for i in d:
                if isinstance(i, str):
                    output.append(os.path.join(parent_key, i))
                else:
                    self.flatten_repo_tree(output, i, parent_key)

        return output

    def exec(self) -> list:
        output = []
        return self.flatten_repo_tree(output, self.paths, self.base_path)



import toml
import os
from typing import Union, MutableMapping

from tackle.models import BaseHook, Field


class TomlHook(BaseHook):
    """
    Hook for reading and writing yaml. Hook reads from `path` if no `data` field is
     provided, otherwise it writes the `data` to `path`.
    """

    hook_type: str = 'toml'
    path: str = Field(..., description="The file path to put read or write to.")
    data: Union[dict, list, str] = Field(
        None,
        description="Map/list or renderable string to a map/list key to write.",
        render_by_default=True,
    )
    args: list = ['path', 'data']

    def exec(self) -> Union[MutableMapping, str]:
        self.path = os.path.abspath(os.path.expanduser(os.path.expandvars(self.path)))
        # Make path if it does not exist
        if not os.path.exists(os.path.dirname(self.path)) and self.data:
            os.makedirs(os.path.dirname(self.path))

        if self.data:
            with open(self.path, 'w') as f:
                toml.dump(self.data, f)
            return self.path

        else:
            data = toml.load(self.path)
            return data



import pytest

from tackle.utils.files import read_config_file
from tackle.exceptions import ContextDecodingException, UnsupportedBaseFileTypeException


def test_read_config_file(change_curdir_fixtures):
    # assert read_config_file('documents.yaml') == [{'this': 'that'}, {'this': 'that'}]
    assert read_config_file('document.yaml')['this'] == 'that'
    assert read_config_file('file.yaml') == {'this': 'that'}
    assert read_config_file('ok.toml')['this'] == {'stuff': 'things'}

    with pytest.raises(ContextDecodingException):
        read_config_file('bad.json')

    with pytest.raises(UnsupportedBaseFileTypeException):
        read_config_file('bad.things')

    with pytest.raises(FileNotFoundError):
        read_config_file('bad')


def test_read_config_file_comments(change_curdir_fixtures):
    document = read_config_file('documents.yaml')
    assert document[0]['this'] == 'that'

    document = read_config_file('document.yaml')
    assert document['this'] == 'that'



from typing import Union, Any

from pydantic import Field

from tackle import BaseHook
from tackle.utils.dicts import encode_key_path, nested_get
from tackle.exceptions import HookCallException

# Hack because fallback value should be able to be None so None can't be default
# -> string instead
FALLBACK_VALUE = "None"


class GetKeyHook(BaseHook):
    """
    Hook for getting a key based on a key path which is a list with keys and numbers
     for indexes in a list.
    """

    hook_type: str = 'get'
    # fmt: off
    path: Union[list, str] = Field(
        ...,
        description="A list or string with a separator for the path to the value you "
                    "want to update with strings for keys and ints for indexes in the "
                    "list."
    )
    sep: str = Field('/', description="For string paths, a separator for key path.")
    fallback: Any = Field(
        FALLBACK_VALUE, description="Fallback value if the keys is not found."
    )
    args: list = ['path', 'fallback']
    # fmt: on

    def exec(self):
        """Run the hook."""
        value = None
        for i in ['public', 'private', 'temporary', 'existing']:
            try:
                value = nested_get(
                    element=getattr(self, f'{i}_context'),
                    keys=encode_key_path(self.path, self.sep),
                )
            except KeyError:
                pass

            if value is not None:
                break

        if value is None and self.verbose:
            if self.verbose:
                print(f"Could not find a key in {self.path} in any context.")
            if self.fallback == FALLBACK_VALUE:
                raise HookCallException(
                    f"Could not find a key in {self.path} in any context.", context=self
                )
            if self.verbose:
                print(f"Using fallback={self.fallback}.")
            value = self.fallback

        return value



from typing import Union, Optional

from tackle import BaseHook, Field
from tackle.utils.dicts import encode_key_path, nested_get, get_target_and_key


class DictKeysHook(BaseHook):
    """Hook for returning the keys of a dict as a list."""

    hook_type: str = 'keys'

    # fmt: off
    src: Union[dict, str, list] = Field(
        description="A dict to get the keys from and output the result or a str with separators or list for a key path to the element to get the keys from within the context.")
    sep: str = Field('/', description="For string src's, a separator for key path.")
    # fmt: on

    args: list = ['src']

    def exec(self) -> Optional[list]:
        """Convert to list then look up before lifting keys off a dict."""
        key_path = None
        if isinstance(self.src, str):
            key_path = encode_key_path(self.src, self.sep)
        elif isinstance(self.src, list):
            key_path = self.src

        if key_path is not None:
            target_context, trim_key_path = get_target_and_key(self, key_path=key_path)

            self.src = nested_get(
                element=target_context,
                keys=trim_key_path,
            )

        return list(self.src.keys())



import re
from jinja2 import meta
from jinja2.exceptions import UndefinedError, TemplateSyntaxError
from inspect import signature
from typing import TYPE_CHECKING, Any, Callable
from pydantic import ValidationError
from pydantic.main import ModelMetaclass

from tackle.hooks import LazyBaseFunction
from tackle.special_vars import special_variables
from tackle.exceptions import (
    UnknownTemplateVariableException,
    MissingTemplateArgsException,
    MalformedTemplateVariableException,
)
from tackle.utils.imports import get_public_or_private_hook
from tackle.utils.command import literal_eval

if TYPE_CHECKING:
    from tackle.models import Context, JinjaHook


def render_variable(context: 'Context', raw: Any):
    """
    Render the raw input. Does recursion with dict and list inputs, otherwise renders
    string.

    :param raw: The value to be rendered.
    :return: The rendered value as literal type.
    """
    if raw is None:
        return None
    elif isinstance(raw, str):
        return render_string(context, raw)
    elif isinstance(raw, dict):
        return {
            render_string(context, k): render_variable(context, v)
            for k, v in raw.items()
        }
    elif isinstance(raw, list):
        return [render_variable(context, v) for v in raw]
    else:
        return raw


def create_jinja_hook(context: 'Context', hook: 'ModelMetaclass') -> 'JinjaHook':
    """Create a jinja hook which is callable via wrapped_exec."""
    from tackle.models import JinjaHook, BaseContext

    return JinjaHook(
        hook=hook,
        context=BaseContext(
            input_context=context.input_context,
            public_context=context.public_context,
            existing_context=context.existing_context,
            no_input=context.no_input,
            calling_directory=context.calling_directory,
            calling_file=context.calling_file,
            public_hooks=context.public_hooks,
            private_hooks=context.private_hooks,
            key_path=context.key_path,
            verbose=context.verbose,
        ),
    )


def add_jinja_hook_methods(
    context: 'Context',
    jinja_hook: 'JinjaHook',
):
    """Recursively look through hook and add methods to jinja hook so that if"""
    from tackle.parser import create_function_model

    for k, v in jinja_hook.hook.__fields__.items():
        if v.type_ == Callable:
            # Enrich the method with the base attributes
            for i in jinja_hook.hook.__fields__['function_fields'].default:
                # Don't override attributes within the method
                if i not in v.default.function_dict:
                    v.default.function_dict[i] = jinja_hook.hook.__fields__[
                        'function_dict'
                    ].default[i]

            # Build the function with a copy of the dict, so it can be called twice
            # without losing methods
            method_base = create_function_model(
                context=context, func_name=k, func_dict=v.default.function_dict.copy()
            )

            # Create the jinja method with
            jinja_method = create_jinja_hook(context, method_base)

            # Add the jinja method's exec as an attribute of the base's wrapped exec so
            # that it can be called within the jinja globals along with the wrapped exec
            jinja_method.set_method(k, jinja_method.wrapped_exec)

            add_jinja_hook_methods(context, jinja_method)


# wrapped_exec calls exec on the `hook` integrating any positional args
def render_string(context: 'Context', raw: str):
    """
    Render strings by first extracting renderable variables then build a render context
    from the public_context, then existing context, and last looks up special variables.
    After the value has been rendered it is returned as literal so as to preserve the
    original type of the value.

    :return: The literal value if the output is a string / list / dict / float / int
    """
    if ('{{' not in raw) and ('{%' not in raw):
        return raw

    try:
        template = context.env_.from_string(raw)
    except TemplateSyntaxError as e:
        raise MalformedTemplateVariableException(
            str(e).capitalize() + f" in {raw}", context=context
        ) from None

    # Extract variables
    variables = meta.find_undeclared_variables(context.env_.parse(raw))

    # We need to make a list of used hooks so that we can remove them from the jinja
    # environment later.
    used_hooks = []

    # Build a render context by inspecting the renderable variables
    render_context = {}
    unknown_variables = []
    for v in variables:
        # Variables in the current public_context take precedence
        if context.temporary_context and v in context.temporary_context:
            render_context.update({v: context.temporary_context[v]})
        elif context.public_context and v in context.public_context:
            render_context.update({v: context.public_context[v]})
        elif context.private_context and v in context.private_context:
            render_context.update({v: context.private_context[v]})
        elif context.existing_context and v in context.existing_context:
            render_context.update({v: context.existing_context[v]})
        elif v in special_variables:
            # If it is a special variable we need to check if the call requires
            # arguments, only context supported now.
            argments = list(signature(special_variables[v]).parameters)
            if len(argments) == 0:
                render_context.update({v: special_variables[v]()})
            elif 'context' in argments:
                render_context.update({v: special_variables[v](context)})
            elif 'kwargs' in argments:
                # TODO: This should support callable special vars
                raise NotImplementedError
            else:
                raise ValueError("This should never happen.")
        else:
            unknown_variables.append(v)

    # Evaluate any hooks and insert into jinja environment globals
    if len(unknown_variables) != 0:
        # Unknown variables can be real unknown variables, preloaded jinja globals or
        # hooks which need to be inserted into the global env so that they can be called
        for i in unknown_variables:
            if i in context.public_hooks or i in context.private_hooks:
                hook = get_public_or_private_hook(context, i)

                # Keep track of the hook put in globals so that it can be removed later
                used_hooks.append(i)

                if isinstance(hook, LazyBaseFunction):
                    # In this case the hook was imported from the `hooks` directory and
                    # it needs to be created before being put in the jinja.globals.
                    from tackle.parser import create_function_model

                    hook = create_function_model(
                        context=context,
                        func_name=i,
                        # Copying allows calling hook twice - TODO: carry over arrow
                        func_dict=hook.function_dict.copy(),
                    )

                    # Replace the provider hooks with instantiated function
                    # context.provider_hooks[i] = hook
                    # TODO: carry over arrow to know where to put hook

                # Create the jinja method with
                jinja_hook = create_jinja_hook(context, hook)

                # Iterate recursively through hook fields looking for methods as
                # type Callables and create attributes on their base function so
                # that those methods can be called individually.
                add_jinja_hook_methods(context, jinja_hook)

                # Add the hook to the jinja environment globals along with all hook
                # methods
                context.env_.globals[i] = jinja_hook.wrapped_exec

            else:
                # An error will be thrown later when the variable can't be rendered
                pass
    try:
        rendered_template = template.render(render_context)

        # Need to remove the hook from the globals as if it is called a second time,
        # it is no longer an unknown variable and will use the prior arguments if they
        # have not been re-instantiated.
        for i in used_hooks:
            context.env_.globals.pop(i)

        # # TODO: RM?
        if rendered_template.startswith('<bound method'):
            # Handle unknown variables that are the same as hook_types issues/55
            raise UndefinedError(
                f"A variable `{'/'.join(used_hooks)}` is the same as "
                f"a hook and either not declared as a variable or "
                f"doesn't have arguments. Consider changing."
            )

        # Check for ambiguous globals like `namespace` tackle-box/issues/19
        match = re.search(r'\<class \'(.+?)\'>', rendered_template)
        if match:
            ambiguous_key = match.group(1).split('.')[-1].lower()
            if ambiguous_key in context.public_context:
                rendered_template = context.public_context[ambiguous_key]
            elif match.group(1) in context.existing_context:
                rendered_template = context.existing_context[ambiguous_key]

    except TypeError as e:
        # Raised when the wrong type is provided to a hook
        # TODO: Consider detecting when StrictUndefined is part of e and raise an
        #  UnknownVariable type of error
        raise UnknownTemplateVariableException(str(e), context=context) from None
    except UndefinedError as e:
        raise UnknownTemplateVariableException(str(e), context=context) from None
    except ValidationError as e:
        # Raised when the wrong type is provided to a hook
        raise MissingTemplateArgsException(str(e), context=context) from None

    # Return the literal type with a few exception handlers
    return literal_eval(rendered_template)



from jinja2.exceptions import UndefinedError
from jinja2 import FileSystemLoader

from typing import Union

from tackle import BaseHook, Field
from tackle.providers.generate.hooks.exceptions import UndefinedVariableInTemplate
from tackle.utils.dicts import get_readable_key_path


class JinjaHook(BaseHook):
    """
    Hook for jinja templates. If given an `output`, the rendered contents are output to
     a file, otherwise the rendered contents are output as a string.
    """

    hook_type: str = 'file_update'
    # fmt: off
    template: str = Field(..., description="Path to the template to render relative to `file_system_loader`.")

    start_marker: str = Field("<!-start->", description="Some marker in the template to start replacing the contents within.")
    end_marker: str = Field("<!-end->", description="Some marker in the template to stop replacing the contents within.")

    start_line: int = Field(None, description="The line number to update from. Supersedes marker.")

    extra_context: dict = Field(None,
                                description="Extra context update the global context to render with.")
    render_context: dict = Field(
        None, description="A render context that invalidates the default context."
    )
    additional_context: dict = Field(
        None, description="A map to use as additional context when rendering."
    )
    file_system_loader: Union[str, list] = Field(
        '.',
        description="List of paths or string path to directory with templates to load "
                    "from. [Docs](https://jinja.palletsprojects.com/en/3.0.x/api/#jinja2.FileSystemLoader).")
    # fmt: on

    # args: list = ['template', 'output']

    _wip = True

    def _init_context(self):
        # Update the render_context that will be used
        if self.render_context is not None:
            return

        # fmt: off
        existing_context = self.existing_context if self.existing_context is not None else {}
        temporary_context = self.temporary_context if self.temporary_context is not None else {}
        private_context = self.private_context if self.private_context is not None else {}
        public_context = self.public_context if self.public_context is not None else {}
        # fmt: on

        self.render_context = {
            **existing_context,
            **temporary_context,
            **private_context,
            **public_context,
        }

        if self.extra_context is not None:
            if isinstance(self.extra_context, list):
                for i in self.extra_context:
                    self.render_context.update(i)
            else:
                self.render_context.update(self.extra_context)

    def find_line_in_template(self, file):
        start_line = None
        end_line = None

        for i, v in enumerate(file):
            if start_line is None and v.startswith(self.start_marker):
                start_line = i + 1
            if (
                start_line is not None
                and end_line is None
                and v.startswith(self.end_marker)
            ):
                end_line = i

        if end_line is None:
            end_line = len(file)

        return start_line, end_line

    def exec(self):
        self._init_context()
        with open(self.template) as f:
            file = f.readlines()

            if self.start_line is None:
                self.start_line, end_line = self.find_line_in_template(file)

            for i in range(self.start_line, end_line):
                file.pop(i)

            print()

        # self.env_.loader = FileSystemLoader(self.file_system_loader)
        # template = self.env_.get_template(self.template)
        #
        # try:
        #     output_from_parsed_template = template.render(**self.render_context)
        # except UndefinedError as err:
        #     msg = f"The Jinja hook for '{get_readable_key_path(self.key_path)}' key path failed to render"
        #     raise UndefinedVariableInTemplate(msg, err, self.public_context)
        #
        # if self.output is not None:
        #     with open(self.output, 'w') as fh:
        #         fh.write(output_from_parsed_template)
        #     return self.output
        # else:
        #     return output_from_parsed_template



"""Collection of tests around loading extensions."""
from tackle.main import tackle
from tackle.models import Context


def test_env_should_come_with_default_extensions():
    """Verify default extensions loaded with StrictEnvironment."""
    c = Context()
    assert 'capitalize' in c.env_.filters
    # TODO: Reassess when adding back filters?
    # assert 'json' in c.env_.filters
    # assert 'get' in c.env_.filters


def test_env_should_evaluate_is_defined(change_curdir_fixtures):
    """Verify that the `is defined` works when rendering."""
    o = tackle('is-defined.yaml', no_input=True)
    assert 'defined' in o
    assert 'not_defined' not in o
    assert 'not_defined_again' not in o
    assert 'defined_again' in o



import pytest

from tackle import exceptions, tackle

EXCEPTIONS = [
    (
        exceptions.UnknownTemplateVariableException,
        'unknown-variable.yaml',
        'is undefined',
    ),
    (
        exceptions.UnknownTemplateVariableException,
        'unknown-hook.yaml',
        'is the same as a hook and either',
    ),
    (
        exceptions.MissingTemplateArgsException,
        'hooks-args-wrong-type.yaml',
        'str type expected',
    ),
    (
        exceptions.MissingTemplateArgsException,
        'hooks-missing-args.yaml',
        '',
    ),
    (
        exceptions.TooManyTemplateArgsException,
        'hooks-args-too-many-args.yaml',
        '',  # TODO: Consider updating to better error
    ),
]


@pytest.mark.parametrize("exception,fixture,error_message", EXCEPTIONS)
def test_render_raise_error(change_curdir_fixtures, exception, fixture, error_message):
    with pytest.raises(exception) as e:
        tackle(fixture)

    assert error_message in e.value.message



import sys
from InquirerPy import prompt
from pprint import pprint
from rich import print

from tackle import BaseHook, Field


class DebugHook(BaseHook):
    """Hook for debugging that prints the output context and pauses runtime."""

    hook_type: str = 'debug'
    key: str = Field(None, description="A path to a key to debug")
    context: str = Field(
        None,
        description="Which context to examine. One of `public`, `private`, "
        "`temporary`, or `existing`. Omit for all.",
    )

    _contexts: list = ['public', 'private', 'temporary', 'existing']

    def print_key(self, print_context):
        if self.key is not None:
            if self.key in print_context:
                pprint(print_context[self.key])
                return True
            else:
                return False
        else:
            pprint(print_context[self.key])

    def print_context(self, print_context, context_name: str):
        print(f"[bold magenta]{context_name.title()} Context[/bold magenta]")

        # TODO: Improve this -> The builtin pprint is better now since it does the
        #  first level as the top level keys.
        # pretty = Pretty(dict(print_context))
        # panel = Panel(pretty)
        # print(panel)
        pprint(print_context)

    def exec(self) -> None:
        if self.context is not None:
            if self.context in self._contexts:
                output = getattr(self, f'{self.context}_context')
                if output is not None:
                    self.print_context(output, self.context)
                else:
                    print(f"Debugging {self.context} not possible because it is empty.")
            else:
                print(
                    f"Input context in debug hook `{self.context}` must be one of "
                    f"`public`, `private`, `temporary`, or `existing`"
                )
        else:
            printed = None
            for i in self._contexts:
                output = getattr(self, f'{i}_context')
                if output is not None and output != {}:
                    if self.key is not None:
                        if self.key in output:
                            self.print_context(output[self.key], i)
                            printed = True
                    else:
                        self.print_context(output, i)
                else:
                    continue

            if self.key is not None and printed is None:
                print(f"Key={self.key} not found in ")

        if not self.no_input:
            question = {
                'type': 'confirm',
                'name': 'tmp',
                'message': 'CONTINUE',
            }
            try:
                response = prompt([question])
            except KeyboardInterrupt:
                print("Exiting...")
                sys.exit(0)

            # Catch keyboard exits with return an empty dict
            if response == {}:
                sys.exit(0)



"""Tests dict input objects for `tackle.providers.pyinquirer.hooks.select` module."""
from tackle.main import tackle


def test_provider_requests_get(change_dir):
    output = tackle('get.yaml', no_input=True)
    assert output['compact']['url'] == output['expanded']['url']


def test_provider_requests_post(change_dir):
    output = tackle('post.yaml', no_input=True)
    assert output['expanded']['json']['stuff'] == "things"
    assert output['expanded']['json'] == output['compact']['json']


def test_provider_requests_put(change_dir):
    output = tackle('put.yaml', no_input=True)
    assert output['expanded']['json']['stuff'] == "things"


def test_provider_requests_patch(change_dir):
    output = tackle('patch.yaml', no_input=True)
    assert output['expanded']['json']['stuff'] == "things"


def test_provider_requests_delete(change_dir):
    output = tackle('delete.yaml', no_input=True)
    assert output['expanded'] == 200



import itertools

from tackle import BaseHook, Field


class ConcatenateHook(BaseHook):
    """Hook for concatenating a list of items."""

    hook_type: str = 'concat'
    # fmt: off
    src: list = Field(
        ..., description="A list to concatenate the items of.", render_by_default=True)
    # fmt: on

    args: list = ['src']

    def exec(self) -> list:
        return list(itertools.chain.from_iterable(self.src))



from rich.console import Console
from rich.markdown import Markdown

from tackle import BaseHook, Field


class MarkdownPrintHook(BaseHook):
    """Hook for printing markdown and returning the output."""

    hook_type: str = 'markdown'
    text: str = Field(..., description="The text to render as markdown.")
    justify: str = Field(
        None, description="Justify value for paragraphs. Defaults to None."
    )

    args: list = ['text']

    # TODO: Map this https://rich.readthedocs.io/en/stable/reference/markdown.html?highlight=markdown%20#rich.markdown.Markdown
    # https://github.com/robcxyz/tackle-box/issues/57
    def exec(self):
        console = Console()
        console.print(Markdown(self.text, justify='left', inline_code_lexer='python'))
        return self.text



import sys

from InquirerPy import prompt
from typing import Any, List, Union

from tackle.models import BaseHook, Field
from tackle.utils.dicts import get_readable_key_path


class InquirerListHook(BaseHook, smart_union=True):
    """
    Hook for PyInquirer 'list' type prompts, a single selector that returns a string.
     Takes in two forms of `choices` inputs, list of string or list of maps with the
     key as the question and the value as the output.
     [Source example](https://github.com/CITGuru/PyInquirer/blob/master/examples/list.py)
    """

    hook_type: str = 'select'

    message: str = Field(None, description="String message to show when prompting.")
    choices: Union[List[str], List[dict]] = Field(
        ...,
        description="List of strings or dicts with keys as output and values as display.",
        render_by_default=True,
    )
    index: bool = Field(
        False, description="Boolean to return the index instead of the answer"
    )

    args: list = ['message']
    _docs_order = 1

    def exec(self) -> Any:
        if self.message is None:
            self.message = get_readable_key_path(self.key_path) + ' >>>'

        # Figure out what type of dictionary it is
        choices_type = None
        for i, v in enumerate(self.choices):
            if i != 0 and type(v) != choices_type:
                raise ValueError("All items need to be of the same type in ")
            choices_type = type(v)

        if choices_type == str:
            answer = self._run_prompt()
            if self.index:
                return self.choices.index(answer)
            else:
                return answer
        elif choices_type == dict:
            choice_list = self.choices.copy()
            self.choices = []
            for i in choice_list:
                self.choices.append(list(i.keys())[0])

            answer = self._run_prompt()
            for i, v in enumerate(choice_list):
                if answer == list(v.keys())[0]:
                    if self.index:
                        return i
                    else:
                        return list(v.values())[0]
        else:
            raise ValueError(
                "Choices must be list of string or dict with displayed "
                "choices as keys and output selection as the value."
            )

    def _run_prompt(self):
        if not self.no_input:
            question = {
                'type': 'list',
                'name': 'tmp',
                'message': self.message,
                'choices': self.choices,
            }

            # Handle keyboard exit
            try:
                response = prompt([question])
            except KeyboardInterrupt:
                print("Exiting...")
                sys.exit(0)
            return response['tmp']

        elif isinstance(self.choices[0], str):
            return self.choices[0]
        elif isinstance(self.choices[0], dict):
            return list(self.choices[0].keys())[0]



from typing import Optional

from tackle import BaseHook, Field


class ListKeyValuesHook(BaseHook):
    """Hook for getting a list of values from a list of maps based on a key."""

    hook_type: str = 'list_key_values'
    # fmt: off
    src: list = Field(
        ..., description="A list to extract the keys out of.", render_by_default=True)
    key: str = Field(...)
    reject_keys: dict = Field(
        None, description="A dict of keys with values to not include in output.")
    src_is_key_path: bool = Field(
        False, description="If the src is a list and is meant to be a key path.")
    sep: str = Field('/', description="For string src's, a separator for key path.")
    # fmt: on

    args: list = ['src', 'key']

    def reject(self, i) -> bool:
        for k, v in self.reject_keys:
            if k in i and i[k] == v:
                return False
        return True

    def exec(self) -> Optional[list]:
        if self.reject_keys is None:
            return [i[self.key] for i in self.src]
        else:
            return [i[self.key] for i in self.src if self.reject(i)]


# class KeyListHook(BaseHook):
#     """Hook for getting an item in a list of maps with a key matching a value."""
#
#     hook_type: str = 'list_key_value'
#     # fmt: off
#     src: Union[list, str] = Field(
#         ..., description="A list to extract the keys out of.")
#     key: str = Field(...)
#     src_is_key_path: bool = Field(
#         False, description="If the src is a list and is meant to be a key path.")
#     sep: str = Field('/', description="For string src's, a separator for key path.")
#     # fmt: on
#
#     args: list = ['src', 'key']
#
#     def exec(self) -> Optional[list]:
#         if isinstance(self.src, list):
#             return [i[self.key] for i in self.src]
#         else:
#             raise NotImplementedError



from tackle import tackle
import datetime


def test_provider_date_date_now(change_dir):
    output = tackle('date_now.yaml')
    assert output['now'] == datetime.datetime.now().strftime("%m-%d-%Y")
    assert output['year'] == datetime.datetime.now().strftime("%Y")
    assert isinstance(output['timestamp'], float)
    assert isinstance(output['timestamp_utc'], float)



from typing import Union

from tackle import BaseHook, Field
from tackle.utils.dicts import encode_key_path, nested_get, get_target_and_key


class DictPopHook(BaseHook, smart_union=True):
    """
    Hook for removing (`pop`) a key from a dict or item from a list based on index.
    Follows python's [pop](https://docs.python.org/3/tutorial/datastructures.html)
    """

    hook_type: str = 'pop'

    # fmt: off
    src: Union[dict, str, list] = Field(
        description="A list to remove an item from or dict to remove a key and output the result or a str with separators or list for a key path to the item operate on within the context.")
    src_is_key_path: bool = Field(False, description="If the src is a list and is meant to be a key path.")
    item: Union[str, int] = Field(
        None,
        description="A string for a key to remove from a dict `src` or integer for an index to remove from a list `src`.",
    )
    sep: str = Field('/', description="For string src's, a separator for key path.")
    # fmt: on

    args: list = ['src', 'item']

    def pop_item(self):
        if self.item is None:
            self.src.pop()
        elif isinstance(self.item, list):
            for i in self.item:
                self.src.pop(i)
        else:
            self.src.pop(self.item)

    def exec(self) -> Union[dict, list, None]:
        """
        Check if the src is a ref to a key in the context or a literal value that will
        be returned.
        """
        if isinstance(self.src, str) or self.src_is_key_path:
            target_context, set_key_path = get_target_and_key(
                self, key_path=encode_key_path(self.src, self.sep)
            )

            self.src = nested_get(
                element=target_context,
                keys=set_key_path,
            )

            # self.src = encode_key_path(self.src, self.sep)
            # self.src = nested_get(
            #     element=self.public_context,
            #     keys=self.src,
            # )
            self.pop_item()
        else:
            self.pop_item()
            return self.src



import sys
import logging
import subprocess
import errno
import struct
import shutil
import os
from itertools import chain
from select import select
from pydantic import Field
from typing import Any

from tackle.models import BaseHook
from tackle.exceptions import HookCallException

if os.name != 'nt':
    # Don't import on windows as pty is not available there
    import pty
    import fcntl
    import termios

logger = logging.getLogger(__name__)


class CommandFailedException(Exception):
    """Exception to stop crazy stack traces everytime an errored command is run."""

    def __init__(self, message=None):
        sys.tracebacklimit = 0
        self.message = message


class CommandHook(BaseHook):
    """Run system commands."""

    hook_type: str = 'command'

    command: str = Field(..., description="A shell command.")
    ignore_error: bool = Field(False, description="Ignore errors.")
    multiline: bool = Field(False, description="Don't automatically breakup lines")

    system: bool = Field(
        False,
        description="Use python's os.system command instead of popen based stream "
        "reader.",
    )

    args: list = ['command']

    def _set_size(self, fd):
        """Found at: https://stackoverflow.com/a/6420070."""
        (
            _COLUMNS,
            _ROWS,
        ) = shutil.get_terminal_size(fallback=(80, 20))
        size = struct.pack("HHHH", _ROWS, _COLUMNS, 0, 0)
        fcntl.ioctl(fd, termios.TIOCSWINSZ, size)

    def exec(self) -> Any:
        # TODO: Fix multi-line calls
        # https://github.com/robcxyz/tackle-box/issues/14
        if self.multiline:
            commands = [self.command.replace('\n', ' ').split()]
        else:
            commands = [i.split() for i in self.command.split('\n')]
            if commands[-1] == []:
                commands.pop()

        data = bytes()
        output = []
        for cmd in commands:
            masters, slaves = zip(pty.openpty(), pty.openpty())
            for fd in chain(masters, slaves):
                self._set_size(fd)

            try:
                with subprocess.Popen(
                    cmd, stdin=slaves[0], stdout=slaves[0], stderr=slaves[1]
                ) as p:
                    for fd in slaves:
                        os.close(fd)  # no input
                        readable = {
                            masters[0]: sys.stdout.buffer,  # store buffers seperately
                            masters[1]: sys.stderr.buffer,
                        }
                    while readable:
                        # TODO: Session is not interactive and fails below when interactive
                        #  prompts are displayed
                        # https://github.com/robcxyz/tackle-box/issues/13
                        # x = select(readable, [], [])[0]
                        for fd in select(readable, [], [])[0]:
                            try:
                                data = os.read(fd, 1024)  # read available
                            except OSError as e:
                                if e.errno != errno.EIO:
                                    raise  # XXX cleanup
                                del readable[fd]  # EIO means EOF on some systems
                            else:
                                if not data:  # EOF
                                    del readable[fd]
                                else:
                                    if fd == masters[0]:  # We caught stdout
                                        # TODO: Interactive prompts
                                        print(data.rstrip().decode('utf-8'))
                                    else:  # We caught stderr
                                        print(data.rstrip().decode('utf-8'))
                                    readable[fd].flush()
                # for fd in masters:
                #     os.close(fd)
                logger.debug("Process exited with %s return code." % p.returncode)
                if p.returncode != 0 and not self.ignore_error:
                    if self.verbose:
                        raise HookCallException(data.decode('utf-8'))
                    else:
                        # Exception already printed
                        raise CommandFailedException()
                else:
                    # return data.decode("utf-8")
                    output.append(data.decode("utf-8"))

            except Exception as e:
                if self.ignore_error:
                    return e
                else:
                    raise e

        for fd in masters:
            os.close(fd)

        return '\n'.join(output)


# class CommandHook(BaseHook):
#     """System commands."""
#
#     hook_type: str = 'command'
#
#     command: str = Field(..., description="A shell command.")
#     ignore_error: bool = Field(False, description="Ignore errors.")
#
#     args: list = ['command']
#
#     def exec(self) -> Any:
#         p = subprocess.Popen(
#             self.command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
#         )
#         output, err = p.communicate()
#
#         if err and not self.ignore_error:
#             raise HookCallException(err.decode('utf-8'))
#         if err and self.ignore_error:
#             return err.decode('utf-8')
#
#         return output.decode("utf-8")



from tackle import tackle


def test_function_supplied_kwargs_param_dict(change_curdir_fixtures):
    """
    Test that we can populate a functions fields with a `kwargs` key as a dict that is
     a dict that will update the field's values.
    """
    output = tackle('supplied-kwargs-param-dict.yaml')
    assert output['kwarg']['bar'] == 'bing'
    assert output['kwarg'] == output['call']


def test_function_supplied_kwargs_param_str(change_curdir_fixtures):
    """
    Test that we can populate a functions fields with a `kwargs` key as a str that is
     a dict that will update the field's values.
    """
    output = tackle('supplied-kwargs-param-str.yaml')
    assert output['kwarg']['bar'] == 'bing'
    assert output['kwarg'] == output['call']


def test_function_supplied_args_param_str(change_curdir_fixtures):
    """Test that we can populate a functions args with an `args` key as str."""
    output = tackle('supplied-args-param-str.yaml')
    assert output['call']['bar'] == 'bing'


def test_function_supplied_args_param_list(change_curdir_fixtures):
    """Test that we can populate a functions args with an `args` key as list."""
    output = tackle('supplied-args-param-list.yaml')
    assert output['call']['bar'] == 'bing bling'



from typing import Union

from tackle.models import BaseHook, Field
from tackle.utils.dicts import nested_delete, encode_key_path, get_target_and_key


class DeleteKeyHook(BaseHook):
    """
    Hook for getting a key based on a key path which is a list with keys and numbers
     for indexes in a list.
    """

    hook_type: str = 'delete'
    # fmt: off
    path: Union[list, str] = Field(
        ..., description="A list or string with a separator for the path to the value "
                         "you want to update with strings for keys and ints for "
                         "indexes in the list.")
    sep: str = Field('/', description="For string paths, a separator for key path.")
    args: list = ['path']
    # fmt: on

    def exec(self) -> None:
        """Run the hook."""
        target_context, set_key_path = get_target_and_key(
            self, key_path=encode_key_path(self.path, self.sep)
        )

        nested_delete(
            element=target_context,
            keys=set_key_path,
        )



import os
import pytest

from tackle import tackle
from tackle import exceptions


def test_function_default_hook_no_context_help(change_curdir_fixtures, capsys):
    """Validate that we can run a default hook's help."""
    with pytest.raises(SystemExit):
        tackle('cli-default-hook-no-context.yaml', 'help')
    out, err = capsys.readouterr()
    assert "usage: tackle" in out
    # with capsys.disabled():
    #     print(out)


def test_function_default_hook_no_context_method_call_help(chdir, capsys):
    """Validate that we can run a hook method's help with visible field."""
    chdir('field-hooks-fixtures')
    with pytest.raises(SystemExit):
        tackle('extends-visible.yaml', 'foo', 'help')
    out, err = capsys.readouterr()
    assert "usage: tackle" in out
    assert "network_info" not in out


def test_function_cli_hook_arg_help(change_curdir_fixtures, capsys):
    """Validate that we can run a hook's with an arg."""
    with pytest.raises(SystemExit):
        tackle('cli-hook-no-context.yaml', 'run', 'help')
    out, err = capsys.readouterr()
    assert "usage: tackle" in out
    # with capsys.disabled():
    #     print(out)


def test_function_cli_hook_arg_help_no_arg(change_curdir_fixtures, capsys):
    """Validate that we can run a hook's with an arg."""
    with pytest.raises(SystemExit):
        tackle('cli-hook-no-context.yaml', 'help')
    out, err = capsys.readouterr()
    assert "usage: tackle" in out
    assert 'foo' not in out  # Taken out with `visible:False` field
    # with capsys.disabled():
    #     print(out)


def test_function_cli_no_default_hook(change_curdir_fixtures, capsys):
    """Check when there is no default hook we only display the available methods."""
    with pytest.raises(SystemExit):
        tackle('cli-no-default-hook.yaml', 'help')
    out, err = capsys.readouterr()
    assert "usage: tackle" in out
    # with capsys.disabled():
    #     print(out)


def test_function_cli_default_hook_embedded(change_curdir_fixtures, capsys):
    """Check when getting help with a method that it works."""
    with pytest.raises(SystemExit):
        tackle('cli-default-hook-embedded.yaml', 'run', 'do', 'help')
    out, err = capsys.readouterr()
    assert "usage: tackle" in out
    # with capsys.disabled():
    #     print(out)


# def test_function_extends_visible(change_curdir_fixtures, capsys):
#     """Check when there is no default hook we only display the available methods."""
#     with pytest.raises(SystemExit):
#         tackle('cli-no-default-hook.yaml', 'help')
#     out, err = capsys.readouterr()
#     assert "usage: tackle" in out


def test_function_cli_tackle_help_no_arg(chdir):
    """Check that when we are in a dir with a default tackle file, we can get help."""
    chdir(os.path.join('fixtures', 'a-tackle'))
    with pytest.raises(SystemExit):
        tackle('help')


def test_function_cli_tackle_default_unknown(change_curdir_fixtures, capsys):
    """
    Check that when the type is not provided because there is some hook for the
     default is unknown.
    """
    with pytest.raises(SystemExit):
        tackle('cli-hook-type-unknown.yaml', 'help')
    out, _ = capsys.readouterr()
    assert "[unknown]" in out


def test_function_cli_tackle_help_with_arg(chdir):
    """
    Check that when we are in a dir with a default tackle file, we can get help when
     calling a declarative hook.
    """
    chdir(os.path.join('fixtures', 'a-tackle'))
    with pytest.raises(SystemExit):
        tackle('stuff', 'help')


def test_function_cli_tackle_arg_error(chdir):
    """Check that when we give a bad key, even with help we get an exception."""
    chdir(os.path.join('fixtures', 'a-tackle'))
    with pytest.raises(exceptions.UnknownArgumentException):
        tackle('bad-key', 'help')



"""
TODO: https://github.com/robcxyz/tackle-box/issues/100
 Should change
"""
import yaml
import re
import os
from pydantic import Field

from typing import Union, Dict, List, Any

from tackle.models import BaseHook
from tackle.utils.dicts import merge


class YamlHook(BaseHook):
    """
    Hook for modifying a yaml in place (ie read, transform, and write back to the file
    in one operation).
    """

    hook_type: str = 'yaml_in_place'

    path: str = Field(..., description="The file path to put read or write to.")
    remove: Union[List, str] = Field(
        None, description="Parameter or regex to remove from list or dict."
    )
    contents: Union[Dict, List] = Field(
        None, description="Supplied dictionary or list to write."
    )
    update: Dict = Field(
        None,
        description="Use the python `update` dict method on `contents` before writing",
    )
    filter: List = Field(None, description="List or string to values to.")
    merge_dict: dict = Field(
        None, description="Dict input that recursively overwrites the `contents`."
    )
    in_place: bool = Field(
        False,
        description="Boolean to read the contents of the `path` and then write after modifications.",
    )
    append_items: Union[Dict, str, List[Any]] = Field(
        None, description="List to append to `append_key` key."
    )
    append_keys: Union[Dict, str, List[Any]] = Field(
        None,
        description="String or list of hierarchical keys to append item to. Defaults",
    )
    mode: str = Field(
        None,
        description="The mode that the file should write. Defaults to write 'w'. See https://docs.python.org/3/library/functions.html#open",
    )
    write: bool = Field(None, description="")

    args: list = ['path', 'contents']

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.path = os.path.expanduser(os.path.expandvars(self.path))

    def _load_contents(self):
        if self.contents:
            # We are writing. Context is provided
            self.write = True
            pass
        if self.in_place:
            # We are modifying in place. Context is read from path
            self.write = True
            with open(self.path, 'r') as f:
                self.contents = yaml.safe_load(f)
        elif not self.contents:
            # We are reading. Contents is read from path
            self.write = False
            mode = self.mode or 'r'
            with open(self.path, mode) as f:
                self.contents = yaml.safe_load(f)

    def _append_each_item(self, append_item):
        if isinstance(self.append_keys, str):
            self.contents[self.append_keys].append(append_item)
        elif isinstance(self.append_keys, list):
            for k in self.append_keys[:-1]:
                self.contents = self.contents.setdefault(k, {})
            self.contents[self.append_keys[-1]].append(append_item)
        else:
            self.contents.append(append_item)

    def _remove_from_contents(self, regex):
        if isinstance(self.contents, list):
            self.contents = [i for i in self.contents if not re.search(regex, i)]
        if isinstance(self.contents, dict):
            for k in list(self.contents.keys()):
                if re.search(regex, k):
                    self.contents.pop(k)

    def _modify_dicts(self):
        if self.remove:
            if isinstance(self.remove, str):
                self._remove_from_contents(self.remove)

            if isinstance(self.remove, list):
                for i in self.remove:
                    self._remove_from_contents(i)

        if self.filter:
            if isinstance(self.contents, dict):
                self.contents = {
                    k: v for (k, v) in self.contents.items() if k in self.filter
                }

        if self.update:
            self.contents.update(self.update)

        if self.merge_dict:
            self.contents = merge(self.contents, self.merge_dict)

        if self.append_items:
            if isinstance(self.append_items, str) or isinstance(
                self.append_items, dict
            ):
                self._append_each_item(self.append_items)
            elif isinstance(self.append_items, list):
                for i in self.append_items:
                    self._append_each_item(i)

    def exec(self) -> Union[str, dict]:
        # Load the path into contents unless it already exists
        self._load_contents()
        # Run all the modifiers
        self._modify_dicts()

        if self.write:
            mode = self.mode or 'w'
            with open(self.path, mode) as f:
                yaml.dump(self.contents, f)
                return self.contents
        else:
            # Read operation, just return contents
            return self.contents



"""Data type hook."""
from typing import Any
from pydantic import Field
from tackle.models import BaseHook


class IntegerHook(BaseHook):
    """Hook for casting a variable to an integer type."""

    hook_type: str = 'int'
    input: Any = Field(..., description="Any variable input.", render_by_default=True)
    args: list = ['input']

    def exec(self) -> int:
        return int(self.input)


class FloatHook(BaseHook):
    """Hook for casting a variable to a float type."""

    hook_type: str = 'float'
    input: Any = Field(..., description="Any variable input.", render_by_default=True)
    args: list = ['input']

    def exec(self) -> float:
        return float(self.input)


class BoolHook(BaseHook):
    """Hook for casting a variable to a bool type."""

    hook_type: str = 'bool'
    input: Any = Field(..., description="Any variable input.", render_by_default=True)
    args: list = ['input']

    def exec(self) -> bool:
        return bool(self.input)


class StrHook(BaseHook):
    """Hook for casting a variable to a string type."""

    hook_type: str = 'str'
    input: Any = Field(..., description="Any variable input.", render_by_default=True)
    args: list = ['input']

    def exec(self) -> str:
        return str(self.input)



"""Prompts used in cloning."""
import os
import sys
from InquirerPy import prompt

from tackle.utils.paths import rmtree


def read_user_yes_no(question, default_value):
    """Ask user yes or no for generic question."""
    question = {
        'type': 'list',
        'name': 'tmp',
        'message': question,
        'default': default_value,
        'choices': ['yes', 'no'],
    }
    return prompt([question])['tmp']


def prompt_and_delete(path, no_input=False):
    """
    Ask user if it's okay to delete the previously-downloaded file/directory.

    If yes, delete it. If no, checks to see if the old version should be
    reused. If yes, it's reused; otherwise, Tackle exits.

    :param path: Previously downloaded zipfile.
    :param no_input: Suppress prompt to delete repo and just delete it.
    :return: True if the content was deleted
    """
    # Suppress prompt if called via API
    if no_input:
        ok_to_delete = True
    else:
        question = (
            "You've downloaded {} before. Is it okay to delete and re-download it?"
        ).format(path)

        ok_to_delete = read_user_yes_no(question, 'no')

    if ok_to_delete:
        if os.path.isdir(path):
            rmtree(path)
        else:
            os.remove(path)
        return True
    else:
        ok_to_reuse = read_user_yes_no(
            "Do you want to re-use the existing version?", 'yes'
        )

        if ok_to_reuse:
            return False

        sys.exit()



from tackle import BaseHook, Field
import psycopg2


class PostgresQuery(BaseHook):
    """
    Hook for running a query against a postgres DB and returning a list of json records
    as the response.
    """

    hook_type = "postgres_query"

    # fmt: off
    query: str = Field(..., description="The query to run.")
    dbname: str = Field("postgres", description="The database name (database is a deprecated alias).")
    user: str = Field("postgres", description="User name used to authenticate.")
    password: str = Field(None, description="Password used to authenticate.")
    host: str = Field(None, description="Database host address (defaults to UNIX socket if not provided).")
    port: int = Field(5432, description="Connection port number (defaults to 5432 if not provided).")
    # fmt: on

    def exec(self) -> list:
        connection = psycopg2.connect(
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.dbname,
        )
        cursor = connection.cursor()
        cursor.execute(self.query)
        response = [
            dict((cursor.description[i][0], value) for i, value in enumerate(row))
            for row in cursor.fetchall()
        ]
        return response



from tackle.main import tackle


def test_provider_tackle_flatten_ls(change_dir):
    output = tackle('ls.yaml', no_input=True)

    assert output['empty']['args'] == '.'
    assert output['all']['args'] == '. --all'
    assert output['almost-all']['args'] == '. --almost-all'
    assert output['arg']['args'] == '.hooks'


def test_provider_tackle_flatten_kubectl(change_dir):
    output = tackle('kubectl.yaml', no_input=True)

    assert output['apply']['args']
    assert output['delete']['args']



from ruamel.yaml import YAML
import os
from typing import Union

from tackle.models import BaseHook, Field


class YamlHook(BaseHook):
    """
    Hook for reading and writing yaml. Hook reads from `path` if no `data` field is
     provided, otherwise it writes the `data` to `path`.
    """

    hook_type: str = 'yaml'
    path: str = Field(..., description="The file path to put read or write to.")
    data: Union[dict, list, str] = Field(
        None,
        description="Map/list or renderable string to a map/list key to write.",
        render_by_default=True,
    )
    args: list = ['path', 'data']

    def exec(self) -> Union[dict, str]:
        self.path = os.path.abspath(os.path.expanduser(os.path.expandvars(self.path)))
        # Make path if it does not exist
        if not os.path.exists(os.path.dirname(self.path)) and self.data:
            os.makedirs(os.path.dirname(self.path))

        yaml = YAML()
        if self.data:
            with open(self.path, 'w') as f:
                yaml.dump(self.data, f)
            return self.path

        else:
            with open(self.path, 'r') as f:
                data = yaml.load(f)

            # TODO: Improve this - https://github.com/robcxyz/tackle-box/issues/56
            import json

            data = json.loads(json.dumps(data))

            return data


class YamlEncodeHook(BaseHook):
    """Hook for converting a dict to a yaml encoded string."""

    hook_type: str = 'yamlencode'
    data: Union[dict, list, str] = Field(
        ...,
        description="Map/list or renderable string to data to convert to yaml string.",
        render_by_default=True,
    )
    args: list = ['data']

    def exec(self) -> Union[dict, str]:
        from io import StringIO

        yaml = YAML()
        options = {}
        string_stream = StringIO()
        yaml.dump(self.data, string_stream, **options)
        output_str = string_stream.getvalue()
        string_stream.close()
        return output_str


class YamlDecodeHook(BaseHook):
    """Hook for decoding a yaml string to a dict."""

    hook_type: str = 'yamldecode'
    data: str = Field(..., description="Yaml string to convert to dict.")
    args: list = ['data']

    def exec(self) -> dict:
        yaml = YAML(typ='safe', pure=True)
        output = yaml.load(self.data)
        return output



import pytest
import os
from ruamel.yaml import YAML

from tackle import tackle
from tackle.exceptions import (
    FunctionCallException,
    HookParseException,
    MalformedFunctionFieldException,
)

FIXTURES = [
    ('call.yaml', 'call-output.yaml'),
    ('return.yaml', 'return-output.yaml'),
]


@pytest.mark.parametrize("fixture,expected_output", FIXTURES)
def test_function_model_extraction(change_curdir_fixtures, fixture, expected_output):
    yaml = YAML()
    with open(expected_output) as f:
        expected_output_dict = yaml.load(f)

    output = tackle(fixture)
    assert output == expected_output_dict


def test_function_no_exec(change_curdir_fixtures):
    """
    Check that when no exec is given that by default the input is returned as is and
    validated.
    """
    output = tackle('no-exec.yaml')

    assert output['no_arg_call']['target'] == 'world'
    assert output['arg_call']['target'] == 'things'


def test_function_field_types(change_curdir_fixtures):
    """Check that field types are respected."""
    output = tackle('field-types.yaml')
    # Based on validation of functions in fixture
    assert output


def test_function_extends(change_curdir_fixtures):
    """Check that we can extend a base function."""
    output = tackle('extends.yaml')
    assert output['t'] == ['hello', 'world']


def test_function_method(change_curdir_fixtures):
    """Check that methods work"""
    output = tackle('method-hook.yaml')
    assert output['t'] == {'in': 'foo', 'bar': 'baz'}
    assert output['t'] == output['jinja_method']
    assert output['jinja_base']['bar'] == 'foo'


def test_function_method_simple(change_curdir_fixtures):
    """Check that we can create a method."""
    output = tackle('method-single.yaml')
    assert output['do'] == {"v": ["Hello", "world!"]}


def test_function_method_embed(change_curdir_fixtures):
    """Check that we can create a method."""
    output = tackle('method-embed.yaml')
    assert output['do']['v'] == ['Hello', 'world!']


def test_function_method_inherit(change_curdir_fixtures):
    """Check that we can create a method."""
    output = tackle('method-inherit.yaml')
    assert output == {"t": "fooo"}


def test_function_method_args(change_curdir_fixtures):
    """Check that we can create a method that takes args."""
    output = tackle('method-args.yaml')
    assert output == {'foo': 'bar'}


def test_function_method_call_from_default(change_curdir_fixtures):
    """
    Check that we can create a method that takes args.
    See https://github.com/robcxyz/tackle-box/issues/99
    """
    output = tackle('method-call-from-default.yaml', target='foo')
    assert output['hi'] == 'Hello foo'


def test_function_method_maintain_context(change_curdir_fixtures):
    """Check a method that carries a context with it from the parent object."""
    output = tackle('method-maintain-context.yaml')
    assert output['do_greeter']['v'] == ['Hello', 'world!']
    assert output['jinja_call']['v'] == ['Hello', 'world!']


def test_function_method_nested(change_curdir_fixtures):
    """Check that we can create a method."""
    output = tackle('method-nested-hook.yaml')
    assert output['jinja_method_home']['destination'] == "earth"
    assert output['jinja_method_home'] == output['t_home']


def test_function_method_override(change_curdir_fixtures):
    """
    Check that when we call methods that attributes are properly overridden if they
     exist in the base.
    """
    output = tackle('method-nested-override.yaml')
    assert output['method_overlap_jinja']['home'] == 'earth'
    assert output['method_overlap_compact']['home'] == 'foo'
    assert output['attribute_override_jinja']['home'] == 'bing'
    assert output['attribute_override_compact']['home'] == 'bing'
    assert output['nested_jinja']['home'] == 'baz'
    assert output['nexted_compact']['home'] == 'baz'


def test_function_import_func_from_hooks_dir(change_curdir_fixtures):
    """Assert that we can call functions from local hooks dir."""
    os.chdir('func-provider')
    o = tackle()
    assert o['compact'] == 'a-default'
    assert o['jinja_extension_default'] == 'a-default'
    assert o['jinja_extension'] == 'things'
    # assert o['jinja_filter'] == 'things'


def test_function_method_no_default(change_curdir_fixtures):
    """Assert that method calls with base fields with no default can be run."""
    o = tackle('method-call-no-default.yaml')
    assert o['compact']['v'] == 'foo'
    assert o['compact'] == o['expanded']
    assert o['jinja_base']['word'] == 'foo'
    assert o['jinja_method']['v'] == 'foo'


# Determine what lists do
# def test_function_list_call(change_curdir_fixtures):
#     """Check what compact hooks do."""
#     output = tackle('list-call.yaml')
#     assert output


# TODO: Build compact hook macro
# def test_function_compact(change_curdir_fixtures):
#     """Check what compact hooks do."""
#     output = tackle('compact.yaml')
#     assert output


def test_function_method_base_validate(change_curdir_fixtures):
    """Check that when a method uses a base attribute, that validation still happens."""
    with pytest.raises(Exception) as e:
        tackle('method-base-validate.yaml')
    assert 'string does not match regex' in e.value.message


EXCEPTION_FIXTURES = [
    # Check that return string not found caught
    ('return-str-not-found.yaml', FunctionCallException),
    # Check that type checking works with no exec method.
    ('no-exec-type-error.yaml', HookParseException),
    # Check args are required when they are not supplied.
    ('field-require.yaml', HookParseException),
    # Check that type is one of literals.
    ('field-bad-type.yaml', MalformedFunctionFieldException),
    # Check that type or default is given.
    ('field-type-or-default.yaml', MalformedFunctionFieldException),
]


@pytest.mark.parametrize("fixture,exception", EXCEPTION_FIXTURES)
def test_function_raises_exceptions(change_curdir_fixtures, fixture, exception):
    with pytest.raises(exception):
        tackle(fixture)


FIELD_TYPE_EXCEPTION_FIXTURES = [
    ('str', 'str'),
    ('dict', 'str'),
    ('list', 'str'),
    ('str', 'type'),
    ('dict', 'type'),
    ('list', 'type'),
    ('str', 'default'),
    ('dict', 'default'),
    ('list', 'default'),
]


@pytest.mark.parametrize("type_,field_input", FIELD_TYPE_EXCEPTION_FIXTURES)
def test_function_raises_exceptions_field_types(chdir, type_, field_input):
    """Check that a validation error is returned for each type of field definition."""
    chdir(os.path.join('fixtures', 'field-type-exceptions'))
    with pytest.raises(HookParseException):
        tackle(f'field-types-{type_}-error-{field_input}.yaml')



# """Test the input source part of the parser."""
# # import pytest
# # import shutil
#
# from tackle import tackle
# # from tackle.cli import main
#
# from tackle.parser import update_source
# from tackle.models import Context
#
# # INPUT_SOURCES = [
# #     ("github.com/robcxyz/tackle-demo", 'my')
# #     # ("robcxyz/full-stack-fastapi-postgresql", 'base-project')
# # ]
# #
# #
# # @pytest.mark.parametrize("input_string,output_dir", INPUT_SOURCES)
# # def test_parser_update_source(change_curdir_fixtures, input_string, output_dir):
# #     """Test various inputs."""
# #     context = Context(input_string=input_string, no_input=True)
# #     update_source(context)
# #     shutil.rmtree(output_dir)
#
#
# def test_parser_find_in_parent(chdir):
#     chdir('fixtures/input-key/child')
#     output = tackle('calling.yaml', find_in_parent=True)
#     assert output
#
#
# def test_parser_update_source_key_to_parent(chdir, mocker):
#     """
#     Test that when an input string is not a file/provider source that we correctly
#     traverse to the nearest tackle file and run a key within it.
#     """
#     chdir('fixtures/input-key/child')
#     context = Context(input_string="do_things", no_input=True)
#     update_source(context)
#     print()



"""Verify hook argument related items."""
import pytest
from tackle.utils.command import unpack_args_kwargs_string
from tackle import tackle

FIXTURES = [
    ("this --if \"expanded == 'that'\"", ["this"], {"if": "expanded == 'that'"}, []),
    (
        "this that --if \"expanded == 'that'\"",
        ["this", 'that'],
        {"if": "expanded == 'that'"},
        [],
    ),
]


@pytest.mark.parametrize("input_string,args,kwargs,flags", FIXTURES)
def test_unpack_input_string(input_string, args, kwargs, flags):
    """Verify the hook arguments."""
    args_out, kwargs_out, flags_out = unpack_args_kwargs_string(input_string)
    assert args_out == args
    assert kwargs_out == kwargs
    assert flags_out == flags


def test_parser_multiple_args(change_curdir_fixtures):
    """Test input args."""
    output = tackle('args.yaml', no_input=True)
    assert output['two_args'] == 'foo bar'
    assert output['three_args'] == 'foo bar baz'


def test_parser_tackle_in_tackle(change_curdir_fixtures):
    """Test input args."""
    output = tackle('outer-tackle.yaml', no_input=True)
    assert output['outer']['foo_items'] == ['bar', 'baz']


def test_parser_tackle_in_tackle_arg(change_curdir_fixtures):
    """Test input args."""
    output = tackle('outer-tackle-arg.yaml', no_input=True)
    assert output['outer']['foo_items'] == ['bar', 'baz']


def test_parser_render_hook_input(change_curdir_fixtures):
    """
    Show that when the first argument is a renderable that you just render that with the
    context and logic.
    """
    output = tackle('var-hook.yaml', no_input=True)
    for k, v in output.items():
        if k.startswith('stuff'):
            assert v == 'things'



from ruamel.yaml import YAML

from tackle.macros import function_field_to_parseable_macro
from tackle.models import Context


def test_function_field_to_parseable_macro(change_curdir_fixtures):
    """Test that the macro is able to expand all the keys properly."""
    yaml = YAML()
    with open('func-inputs-basic.yaml') as f:
        fixture = yaml.load(f)

    output = function_field_to_parseable_macro(
        fixture['input'], context=Context(), func_name='f'
    )

    assert fixture['expected'] == output



import os

from tackle.models import BaseHook, Field


class SymlinkHook(BaseHook):
    """
    Hook creating symlinks wrapping `os.symlink` functionality. Wraps
    [`os.symlink`](https://www.geeksforgeeks.org/python-os-symlink-method/)
    """

    hook_type: str = 'symlink'
    src: str = Field(
        ..., description="String or list of sources, either a directories or files"
    )
    dst: str = Field(..., description="String for path to copy to")
    target_is_directory: bool = Field(
        False,
        description="The default value of this parameter is False. If the "
        "specified target path is directory then its value should "
        "be True.",
    )
    overwrite: bool = Field(False, description="Overwrite the destination.")

    args: list = ['src', 'dst']
    _docs_order = 3

    def exec(self) -> None:
        self.src = os.path.abspath(os.path.expanduser(os.path.expandvars(self.src)))
        self.dst = os.path.abspath(os.path.expanduser(os.path.expandvars(self.dst)))

        if self.overwrite and os.path.islink(self.dst):
            os.symlink(
                self.src,
                ''.join([self.dst, '.tmp']),
                target_is_directory=self.target_is_directory,
            )
            os.rename(self.dst + '.tmp', self.dst)
        elif self.overwrite and os.path.isfile(self.dst):
            os.remove(self.dst)
            os.symlink(
                src=self.src, dst=self.dst, target_is_directory=self.target_is_directory
            )
        else:
            os.symlink(
                src=self.src, dst=self.dst, target_is_directory=self.target_is_directory
            )



"""Tests dict input objects for `tackle.providers.tackle.block` module."""
from tackle.main import tackle


def test_provider_system_hook_debug(change_dir):
    output = tackle(no_input=True)

    assert output['t'] is None
    assert 'g' not in output



# from tackle.models import BaseHook, Field
#
#
# class SourceTypeHooks(BaseHook):
#     """Convert git source to/from https to ssh types."""
#
#     hook_type: str = "git_repository"
#
#     url: str = Field(..., description="The git URL, one of https, git, ssh.")
#
#     type: str = None
#
#     args: list = ['url']
#
#     def exec(self):
#         pass
#



import sys
from pydantic.main import ModelMetaclass, ModelField, ValidationError
from pydantic import BaseModel
from jinja2 import Template
from typing import Any, List

from tackle import exceptions
from tackle.models import Context
from tackle.hooks import LazyBaseFunction

HELP_TEMPLATE = """usage: tackle {{input_string}} {% for i in general_kwargs %}{{i}} {% endfor %}
{% if general_help %}
{{general_help}}
{% endif %}{% if flags != [] %}
flags:{% for i in flags %}
    --{{i.name}}      {% if i.description != None %}{{ ('' ~ i.description) | wordwrap(78) }}{% endif %}{% endfor %}{% endif %}{% if kwargs != [] %}
options:{% for i in kwargs %}
    --{{i.name}} [{{ i.type }}] {% if i.description != None %}{{ ('' ~ i.description) | wordwrap(78) }}{% endif %}{% endfor %}{% endif %}{% if args != [] %}
positional arguments:{% for i in args %}
    {{i.name}} [{{ i.type }}] {% if i.description != None %}{{ ('' ~ i.description) | wordwrap(78) }}{% endif %}{% endfor %}{% endif %}{% if methods != [] %}
methods:{% for i in methods %}
    {{i.name}}     {% if i.description != None %}{{ ('' ~ i.description) | wordwrap(78) }}{% endif %}{% endfor %}{% endif %}
"""


class HelpInput(BaseModel):
    """Validate help input."""

    name: str
    type: str
    default: Any = None
    description: str = None


def unpack_hook(
    context: 'Context', hook: ModelMetaclass
) -> (List[dict], List[dict], List[dict], List[dict]):
    """Unpack arguments (args/kwargs/flags/methods) from hook."""
    args = []
    kwargs = []
    flags = []
    methods = []

    # kwargs + flags
    for i in hook.__fields__['function_fields'].default.copy():
        # Consume the arg as what is left will be a method which will deal with later
        hook_field = hook.__fields__['function_dict'].default.pop(i)

        # Skip fields that have a `visible=False` field
        if isinstance(hook_field, ModelField):
            # Methods fields are of type ModelField
            # TODO: Determine this logic
            pass
        else:
            if 'visible' in hook_field and not hook_field['visible']:
                continue

        # Serialize help input
        if isinstance(hook_field, str):
            help_arg = HelpInput(
                name=i, type=type(hook_field).__name__, default=hook_field
            )
        elif isinstance(hook_field, dict):
            if 'type' not in hook_field:
                hook_field['type'] = "unknown"
            try:
                help_arg = HelpInput(name=i, **hook_field)
            except ValidationError as e:
                hook_name = hook.identifier.split('.')[-1]
                if hook_name == '':  # Default hook
                    hook_name = 'default hook'
                raise exceptions.MalformedFunctionFieldException(
                    f"The field '{i}' is malformed. Getting error=\n{e.__str__()}",
                    context=context,
                    function_name=hook_name,
                ) from None
        elif isinstance(hook_field, ModelField):
            # TODO: https://github.com/robcxyz/tackle-box/issues/91
            # If enabling inheritance for base vars in to the help, then this will
            # neeed to be active.
            # help_arg = HelpInput(
            #     name=i,
            #     type=hook_field.type_.__name__,
            #     # description=hook_field.d  # TODO: Fix this
            # )
            continue
        else:
            raise NotImplementedError

        if help_arg.type == 'bool':
            flags.append(help_arg.dict())
        else:
            kwargs.append(help_arg.dict())

    # args
    for arg in hook.__fields__['args'].default:
        # Check kwargs
        arg_data = [i for i in kwargs if i['name'] == arg]
        if len(arg_data) == 1:
            arg_data = arg_data[0]
        else:
            # Check flags
            arg_data = [i for i in flags if i['name'] == arg]
            if len(arg_data) == 1:
                arg_data = arg_data[0]
            else:
                raise exceptions.MalformedFunctionFieldException(
                    "The args don't match a key word arg or flag.",
                    # TODO: Fix
                    function_name=hook.__fields__,
                    context=context,
                ) from None
        args.append(arg_data)

    # methods
    for k, v in hook.__fields__['function_dict'].default.items():
        if k.endswith('<-') and k not in ['exec', 'exec<-']:
            methods.append(
                HelpInput(
                    name=k[:-2],
                    type='method',
                    description=v['help'] if 'help' in v else None,
                ).dict()
            )

    return args, kwargs, flags, methods


def get_methods_on_default_hook(context: 'Context') -> List[dict]:
    """
    When showing help for the default hook, we must show the additional methods. This
    function gets those methods from the context.
    """
    methods = []

    for k, v in context.public_hooks.items():
        # Handle hooks that could be in the `hooks` dir and will not be instantiated
        if isinstance(v, LazyBaseFunction):
            method_help = ''
            if 'help' in v.function_dict:
                method_help = v.function_dict['help']
        else:
            # Normal hook which has been instantiated
            method_help = (
                v.__fields__['function_dict'].default['help']
                if 'help' in v.__fields__['function_dict'].default
                else ''
            )

        methods.append(
            HelpInput(
                name=k,
                type='method',
                description=method_help,
            ).dict()
        )
    return methods


def run_help(context: 'Context', hook: ModelMetaclass = None):
    """
    Print the help screen then exit. Help can be displayed in three different scenarios.
    1. For the base (no args) -> this shows default / other function's help
    2. For a function (len(args) = 1) -> this shows function's help along with its
        associated methods. This is recursive so depends on depth of methods.
    3. For a function's methods (len(args) > 1) -> this drills into methods
    """
    if hook is not None:
        # We have a default hook or method
        general_help = (
            hook.__fields__['function_dict'].default.pop('help')
            if 'help' in hook.__fields__['function_dict'].default
            else None
        )

        args, kwargs, flags, methods = unpack_hook(context, hook)
        hook_name = hook.identifier.split('.')[-1]
    else:
        # Situation where we don't have a default hook or we have not been given an arg
        # to call a specific hook and need to just display all the base's methods
        args, kwargs, flags, methods, general_help = [], [], [], [], ""
        hook_name = ''

    if hook_name == '':  # Default hook
        hook_name = '_default_'
        # Add the additional methods adjacent to the default hook.
        methods += get_methods_on_default_hook(context=context)

    general_kwargs = []
    for i in kwargs + flags:
        # Show the other options outside the default hook
        general_kwargs.append(f"[--{i['name']}]")

    # Remove the `help` str from the `usage` line in help.
    if context.input_string[-4:] == 'help':
        context.input_string = context.input_string[:-4]

    template = Template(HELP_TEMPLATE)
    help_rendered = template.render(
        args=args,
        kwargs=kwargs,
        flags=flags,
        general_help=general_help,
        general_kwargs=general_kwargs,
        input_string=context.input_string,
        methods=methods,
        hook_name=hook_name,
    )
    print(help_rendered)
    sys.exit(0)



from tackle import tackle


def test_provider_system_hook_dicts_update(change_dir):
    output = tackle('update.yaml')
    assert output['update_map'] == output['arg']
    assert output['update_map2'] == output['arg2']
    assert output['dict_map'] == {'foo': 'bar', 'bar': 'bar'}
    assert output['input_map']['foo'] == 'bing'
    assert output['new_list'][2] == 'baz'
    assert output['new_str'] == 'baz'


def test_provider_system_hook_dicts_update_2(change_dir):
    output = tackle('update-2.yaml')
    assert output


def test_provider_system_hook_dicts_pop(change_dir):
    output = tackle('pop.yaml')
    assert 'stuff' not in output['pop_map']
    assert output['arg_1'] == ['stuff']
    assert 'foo' not in output['arg_2']
    assert 'baz' in output['arg_2']
    assert output['embedded']['list'] == ['stuff']
    assert output['list'] == ['things']


def test_provider_system_hook_dicts_keys(change_dir):
    """Validated keys hook outputs list from map."""
    output = tackle('keys.yaml')
    # Validated with inline assertions
    assert output['output'] == ['stuff', 'things']
    assert output['check']
    assert output['check_key_path']


def test_provider_system_hook_dicts_values(change_dir):
    """Validated keys hook outputs list from map."""
    output = tackle('values.yaml')
    # Validated with inline assertions
    assert output['output']
    assert output['check']
    assert output['check_key_path']


def test_provider_context_set(change_dir):
    """Check that we can set keys."""
    output = tackle('set.yaml')
    assert output['one'][0]['that']['stuff'] == 'more things'
    assert output['two'][0]['that']['stuff'] == 'more things'
    assert output['three'][0]['that']['stuff'] == 'more things'


def test_provider_context_get(change_dir):
    """Check that we can get keys."""
    output = tackle('get.yaml')
    assert output['getter_list'] == 'things'
    assert output['getter_str'] == 'things'
    assert output['getter_str_sep'] == 'things'


def test_provider_context_delete(change_dir):
    """Check that we can delete keys."""
    output = tackle('delete.yaml')
    assert output['one'][0]['that'] == {}
    assert output['two'][0]['that'] == {}
    assert output['three'][0]['that'] == {}


def test_provider_context_append(change_dir):
    """Verify the append hook literal and in place."""
    output = tackle('append.yaml')
    assert 'donkey' in output['appended_list']
    expected_output = ['stuff', 'things', 'foo']
    assert output['output'] == expected_output
    assert output['path']['to']['list'] == expected_output
    assert len(output['stuff']) == 2


def test_provider_context_append_dict(change_dir):
    """Verify the append hook literal and in place."""
    output = tackle('append-dict.yaml')
    assert output



import shutil

import pytest
import os
import subprocess
import sys

from tackle.main import tackle
from tackle.settings import settings


@pytest.fixture()
def temporary_uninstall():
    """Fixture to uninstall a package and install it after the test."""

    def f(package):
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "uninstall",
                "--quiet",
                "--disable-pip-version-check",
                "-y",
                package,
            ]
        )

    return f


def test_parser_provider_import_installs_requirements(
    change_curdir_fixtures, temporary_uninstall
):
    """Validate that if a package is missing, that it will be installed and usable."""
    temporary_uninstall('requests')
    try:
        import requests

        # Fails in CI - I believe requests is available from system python as locally
        # this assert works.
        # assert False
    except ImportError:
        assert True

    tackle('test-install-dep.yaml')
    import requests  # noqa

    assert requests


def test_parser_provider_import_requirements_installs_requirements(
    chdir, temporary_uninstall
):
    """Validate that if a package is missing, that it will be installed and usable."""
    chdir(os.path.join('fixtures', 'test-provider-reqs'))
    temporary_uninstall('art')
    try:
        import art

        assert False
    except ImportError:
        assert True

    o = tackle()
    temporary_uninstall('art')
    assert o['art']


def test_parser_hooks_raises_error_on_unknown_hook_type(change_curdir_fixtures):
    """Verify raising error.

    Verify that the hook parser raises the right error when the hook type is
    not in the providers.
    """
    from tackle.exceptions import UnknownHookTypeException

    with pytest.raises(UnknownHookTypeException):
        tackle('unknown-hook-type.yaml')


def test_parser_provider_hook_add(change_curdir_fixtures):
    """Validate adding providers.

    Validate that you can give a `__provider` key to point to
    additional providers and make them available as a type.
    """
    os.chdir('test-provider')
    o = tackle('context_provider.yaml')
    assert o['things'] == 'bar'


def test_parser_provider_hook_add_list(change_curdir_fixtures):
    """Validate that hooks are imported when called from the same directory."""
    o = tackle('context-provider-list.yaml')
    assert o['things'] == 'bar'
    assert o['stuff'] == 'bar'


@pytest.mark.parametrize("input_file", [None, "file.yaml"])
def test_providers_local_hooks_dir(chdir_fixture, input_file):
    """
    Check that when you call from a child dir that you import the hooks from the dir
     with the file being called. Tests both the case that the parent file is a
     tackle.yaml file and that when you import the file via a tackle hook, that the
     hooks are imported and usable.
    """
    chdir_fixture(os.path.join('child', 'dir'))
    o = tackle(input_file)
    assert o['do'] == 'foo'


@pytest.fixture()
def remove_provider():
    """Fixture to remove a provider."""

    def f(provider):
        provider = provider.split('/')
        provider_path = os.path.join(settings.provider_dir, *provider)
        if os.path.exists(provider_path):
            shutil.rmtree(provider_path)

    return f


def test_providers_released(chdir_fixture, remove_provider):
    """
    Check that when we call a released provider, that we only use commits from the
    latest release and not anything that was added after the release. Check the external
     fixture for details.
    """
    remove_provider("robcxyz/tackle-fixture-released")
    o = tackle("robcxyz/tackle-fixture-released")
    assert 'released_added_later' not in o
    assert o['released_hook'] == 'foo'


def test_providers_released_latest(chdir_fixture, remove_provider):
    """
    Check that when we call a released provider, that when we include the latest flag
    we use the latest commit.
    """
    remove_provider("robcxyz/tackle-fixture-released")
    o = tackle(
        "robcxyz/tackle-fixture-released",
        latest=True,
    )
    assert 'released_added_later' in o
    assert o['released_hook'] == 'foo'

    # Then test that when we run the provider again that it uses the latest release.
    o = tackle("robcxyz/tackle-fixture-released")
    assert 'released_added_later' not in o
    assert o['released_hook'] == 'foo'


def test_providers_unreleased_import(chdir_fixture, remove_provider):
    """Check that we can use an unreleased provider."""
    remove_provider("robcxyz/tackle-fixture-unreleased")
    o = tackle("robcxyz/tackle-fixture-unreleased", no_input=True)
    assert o['this'] == 'that'


def test_providers_hook_dirs(change_dir):
    """Check that we can import hooks by supplying a directory."""
    o = tackle("hook-dirs.yaml", hook_dirs=[str(os.path.join('fixtures', '.hooks'))])
    assert o['t'] == 'things'



from typing import Type, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from tackle.models import BaseHook, Context
    from tackle.hooks import LazyBaseFunction


def get_public_or_private_hook(
    context: 'Context',
    hook_type: str,
) -> Union[Type['BaseHook'], 'LazyBaseFunction']:
    """Get the public or private hook from the context."""
    h = context.public_hooks.get(hook_type, None)
    if h is not None:
        return h
    return context.private_hooks.get(hook_type, None)



"""Settings initializer."""
from pydantic import BaseSettings, Field
from typing import Dict
import os
import logging
import pathlib
from xdg import xdg_config_home  # When adding replay functionality -> xdg_state_home

from tackle.utils.files import read_config_file

logger = logging.getLogger(__name__)


class Settings(BaseSettings):
    """Base settings that are immutable during main runtime."""

    tackle_dir: str = os.path.join(xdg_config_home(), 'tackle')
    config_path: str = os.path.join(tackle_dir, 'settings.yaml')
    provider_dir: str = os.path.join(tackle_dir, 'providers')

    extra_providers: list = Field(
        None, description="Extra providers to import into each context."
    )
    default_tackle_file: str = Field(
        None, description="If set, this will be the file if no argument is provided."
    )

    class Config:
        env_prefix = 'TACKLE_'
        env_file_encoding = 'utf-8'
        case_sensitive = False
        extra = "ignore"

    def __init__(self, **values: Dict):
        super().__init__(**values)
        if not os.path.isdir(self.provider_dir):
            pathlib.Path(self.provider_dir).mkdir(parents=True, exist_ok=True)

        if os.path.exists(self.config_path):
            global_settings = read_config_file(self.config_path)
            for k, v in global_settings:
                setattr(self, k, v)


settings = Settings()



import os
import pytest
import json
from ruamel.yaml import YAML

from tackle.cli import main

FIXTURES = [
    ('cli-call-func.yaml', 'cli-call-func-output.yaml', 'func_a'),
]


@pytest.mark.parametrize("fixture,expected_output,argument", FIXTURES)
def test_function_model_extraction(
    change_curdir_fixtures, fixture, expected_output, argument, capsys
):
    main([fixture, argument, "-p"])
    output = capsys.readouterr().out

    yaml = YAML()
    with open(expected_output) as f:
        expected_output = yaml.load(f)

    assert json.loads(output) == expected_output


@pytest.mark.parametrize("fixture,expected_output,argument", FIXTURES)
def test_function_model_extraction_in_directory(
    change_curdir_fixtures, chdir, fixture, expected_output, argument, capsys
):
    yaml = YAML()
    with open(expected_output) as f:
        expected_output = yaml.load(f)

    os.chdir("func-provider")
    main([fixture, argument, "-p", "--find-in-parent"])
    output = capsys.readouterr().out

    assert json.loads(output) == expected_output



import base64

from tackle.models import BaseHook, Field


class Base64EncodeHook(BaseHook):
    """Hook for `base64_encode`. Base64 encodes a string."""

    hook_type: str = 'base64_encode'
    input: str = Field(..., description="A string to encode.")

    args: list = ['input']
    _docs_order = 8

    def exec(self) -> str:
        sample_string_bytes = self.input.encode("utf-8")
        base64_bytes = base64.b64encode(sample_string_bytes)
        base64_string = base64_bytes.decode("utf-8")
        return base64_string


class Base64DecodeHook(BaseHook):
    """Hook for `base64_decode`. Base64 decodes a string."""

    hook_type: str = 'base64_decode'
    input: str = Field(..., description="A string to decode.")

    args: list = ['input']
    _docs_order = 9

    def exec(self) -> str:
        return str(base64.b64decode(self.input).decode("utf-8"))



import os.path
import fnmatch
from typing import Union
from pydantic import Field
from pathlib import Path
from jinja2 import FileSystemLoader
from jinja2.exceptions import UndefinedError
import shutil
from typing import List

from tackle.models import BaseHook
from tackle.providers.generate.hooks.exceptions import (
    UndefinedVariableInTemplate,
    GenerateHookTemplateNotFound,
)


class GenerateHook(BaseHook, smart_union=True):
    """
    Hook for generating project outputs. Recursively renders all files and folders in a
     given target directory to an output. If there is a "templates" directory and a file
     or directory that matches the `templates` input param, use that as target.
    """

    hook_type: str = 'generate'
    # fmt: off
    templates: Union[str, list] = Field(
        "templates",
        description="Path or list of paths to a templatable directory or file to "
                    "recursively render the contents.",
    )
    output: str = Field('.', description="Path to put the output file(s).")
    copy_without_render: Union[str, list] = Field(
        [], description="List of path to files to only copy and not render."
    )
    overwrite_if_exists: bool = Field(
        False, description="Overwrite the output if exists."
    )
    skip_if_file_exists: bool = Field(
        False, description="Skip creating if path exists."
    )
    skip_overwrite_files: list = Field(
        [], description="List of files to skip generating over if they exist."
    )
    render_context: dict = Field(
        None, description="A render context that invalidates the default context."
    )
    extra_context: Union[str, dict, List[dict]] = Field(
        None,
        description="A map / list of maps to use as extra context when rendering. "
                    "Lists inputs are merged together as lists themselves don't make "
                    "sense.",
        render_by_default=True,
    )
    file_system_loader: Union[str, list] = Field(
        '.',
        description="List of paths or string path to directory with templates to load "
                    "from. [Docs](https://jinja.palletsprojects.com/en/3.0.x/api/#jinja2.FileSystemLoader)."  # noqa
    )

    base_dir_: Path = None
    # env_: Any = None
    file_path_separator_: str = None  # / for mac / linux - \ for win
    # fmt: on

    args: list = ['templates', 'output']

    def _init_paths(self):
        if isinstance(self.copy_without_render, str):
            self.copy_without_render = [self.copy_without_render]

        self.output = os.path.expanduser(os.path.expandvars(self.output))
        if 'nt' in os.name:
            self.file_path_separator_ = '\\'
            if not self.output.startswith('\\'):
                self.output = os.path.join(self.calling_directory, self.output)
        else:
            self.file_path_separator_ = '/'
            if not self.output.startswith('/'):
                self.output = os.path.join(self.calling_directory, self.output)

    def _init_context(self):
        # Update the render_context that will be used
        if self.render_context is not None:
            return

        # fmt: off
        existing_context = self.existing_context if self.existing_context is not None else {}
        temporary_context = self.temporary_context if self.temporary_context is not None else {}
        private_context = self.private_context if self.private_context is not None else {}
        public_context = self.public_context if self.public_context is not None else {}
        # fmt: on

        self.render_context = {
            **existing_context,
            **temporary_context,
            **private_context,
            **public_context,
        }

        if self.extra_context is not None:
            if isinstance(self.extra_context, list):
                for i in self.extra_context:
                    self.render_context.update(i)
            else:
                self.render_context.update(self.extra_context)

    def exec(self):
        """Generate files / directories."""
        self._init_paths()
        self._init_context()

        # https://stackoverflow.com/questions/42368678/jinja-environment-is-not-supporting-absolute-paths
        # Need to add root to support absolute paths
        if isinstance(self.file_system_loader, str):
            self.env_.loader = FileSystemLoader([self.file_system_loader, '/'])
        else:
            self.env_.loader = FileSystemLoader(self.file_system_loader + ['/'])

        if isinstance(self.templates, str):
            self.generate_target(self.templates)
        elif isinstance(self.templates, list):
            for target in self.templates:
                self.generate_target(target)

    def generate_target(self, target: str):
        """
        Generate from an unknown target. If there is a `templates` directory in the
        provider, check if the target matches and use that otherwise search for target.
        Generates files or directories from there.

        :param target: A generic target to generate from, file or directory.
        """
        # Update the input to default to searching in `templates` directory
        if os.path.exists(target):
            self.base_dir_ = Path(".")
        elif os.path.exists(os.path.join("templates", target)):
            self.base_dir_ = Path("templates")
        else:
            self.base_dir_ = Path(target).parent.absolute()

        target_path = os.path.join(self.base_dir_, target)

        # Expand the target path
        self.output = os.path.expanduser(os.path.expandvars(self.output))
        if self.output == '.':
            self.output = os.path.basename(target)

        if os.path.isfile(target_path):
            self.generate_file(target_path, self.output)

        elif os.path.isdir(target_path):
            self.generate_dir(target_path, self.output)
        else:
            raise GenerateHookTemplateNotFound(
                f"Could not find {target_path}.", hook=self
            ) from None

    def generate_file(self, input_file: str, output_path: str):
        """
        Take a target input_file and render its contents / file name to an output path.

        :param input_file: Input file to generate from
        :param output_path: Output file to generate to
        """
        if output_path.endswith(self.file_path_separator_):
            output_path = os.path.join(output_path, os.path.basename(input_file))

        # For skip operations -> Useful when we only want to render once
        if os.path.exists(output_path):
            if self.skip_if_file_exists:
                return

            if self.skip_overwrite_files is not None:
                for i in self.skip_overwrite_files:
                    skip_output_path = os.path.join(Path(output_path).parent, i)
                    if output_path == skip_output_path:
                        return

        # Render the path right away as templating mangles things later - also logical
        # to render file names.  Who wants to generate files with templates in the name?
        file_name_template = self.env_.from_string(str(output_path))
        try:
            output_path = file_name_template.render(self.render_context)
        except UndefinedError as e:
            msg = f"The `generate` hook failed to render -> {e}"
            raise UndefinedVariableInTemplate(msg, hook=self) from None

        # Make the parent directories by default
        parent_dir = Path(output_path).parent.absolute()
        Path(parent_dir).mkdir(parents=True, exist_ok=True)

        # Evaluate if it is a copy without render
        if self.is_copy_only_path(input_file):
            shutil.copyfile(input_file, output_path)
            return

        try:
            file_contents_template = self.env_.get_template(os.path.abspath(input_file))
        except UnicodeDecodeError:
            # Catch binary files with this hack and copy them over
            # TODO: Perhaps improve? In cookiecutter they used a package binary-or-not
            # or something like that but we are staying lean on dependencies in this
            # project.
            shutil.copyfile(input_file, output_path)
            return

        try:
            rendered_contents = file_contents_template.render(self.render_context)
        except UndefinedError as e:
            msg = f"The `generate` hook failed to render -> {e}"
            raise UndefinedVariableInTemplate(msg, hook=self) from None

        # Write contents
        with open(output_path, 'w') as f:
            f.write(rendered_contents)

    def generate_dir(self, input_directory: str, output_path: str):

        for i in os.listdir(input_directory):
            input = os.path.join(input_directory, i)
            output = os.path.join(output_path, i)
            if os.path.isdir(input):
                # Path(output).mkdir(parents=True, exist_ok=True)
                self.generate_dir(input, output)

            if os.path.isfile(input):
                self.generate_file(input, output)

    def is_copy_only_path(self, path):
        """Check whether the given `path` should only be copied and not rendered.

        Returns True if `path` matches a pattern in the given `context` dict,
        otherwise False.

        :param path: A file-system path referring to a file or dir that
            should be rendered or just copied.
        """
        if self.copy_without_render is None:
            return False

        for dont_render in self.copy_without_render:
            if fnmatch.fnmatch(path, dont_render):
                return True
            # TODO: Make this more logical - cookiecutter allowed without `./`
            if fnmatch.fnmatch(path, f'.{self.file_path_separator_}' + dont_render):
                return True
        return False



import os
from tackle.main import tackle
import pytest
import platform


@pytest.fixture()
def tmp_env_var():
    """Set env var temporarily."""
    os.environ['TMP_VAR_THING'] = 'stuff'
    yield
    vars = ['TMP_VAR_THING', 'TMP_VAR_STUFF']
    for v in vars:
        if v in os.environ:
            del os.environ[v]


def test_provider_env_hook_envs(change_dir, tmp_env_var):
    if platform.system() == 'Windows':
        # Need to support this later
        pytest.skip()

    o = tackle('envs.yaml')
    assert o['get_env'] == o['get_env_arg'] == o['get_env_arg_fallback']
    assert o['set_env'] == o['set_env_arg']
    assert o['get_env_arg_after_set'] == 'things'
    assert os.environ['TMP_VAR_THING']



import pytest
import os
from tackle.main import tackle


def test_provider_toml_hook_read(change_dir):
    o = tackle('read.yaml', no_input=True)
    assert 'owner' in o['read'].keys()


@pytest.fixture()
def clean_toml():
    """Remove outputs."""
    yield
    if os.path.exists('writing.toml'):
        os.remove('writing.toml')


def test_provider_toml_hook_write(change_dir, clean_toml):
    tackle('write.yaml', no_input=True)
    assert os.path.exists('writing.toml')



"""Test tackle.utils.vcs."""
import os
import pytest

from tackle.settings import settings
from tackle.utils.vcs import (
    parse_repo_ref,
    get_repo_version,
    get_default_branch,
    get_repo_source,
    # clone_to_dir,
    get_latest_release,
)

DEMO_INPUTS = [
    "https://github.com/robcxyz/tackle-demos",
    "github.com/robcxyz/tackle-demos",
    "robcxyz/tackle-demos",
]


@pytest.mark.parametrize("repo_input", DEMO_INPUTS)
def test_parse_repo_ref(repo_input):
    """Test parsing the repo reference."""
    repo_url, organization, provider_name = parse_repo_ref(repo_input)
    assert repo_url == "https://github.com/robcxyz/tackle-demos"
    assert organization == "robcxyz"
    assert provider_name == "tackle-demos"


DEMO_INPUTS = [
    ("https://github.com/robcxyz/tackle-demos", "main", "main"),
    ("github.com/robcxyz/tackle-demos", "latest", "main"),
    (
        "robcxyz/tackle-demos",
        "v0.0.1-alpha.1",
        "e11248f29c7e4ebcd04164be230f41513b576ffe",
    ),
]


@pytest.mark.parametrize("repo_input,repo_version,expected_version", DEMO_INPUTS)
def test_get_repo_source(
    tmp_move_tackle_dir, repo_input, repo_version, expected_version
):
    """Test getting the repo and cloning it to the providers dir returning the path."""
    get_repo_source(repo=repo_input, repo_version=repo_version)
    assert os.listdir(settings.provider_dir) == ['robcxyz']
    assert (
        get_repo_version(os.path.join(settings.provider_dir, 'robcxyz', 'tackle-demos'))
        == expected_version
    )


def test_get_repo_version(change_dir_base):
    """Test get_repo_version."""
    version = get_repo_version('.')
    assert len(version) > 3


def test_get_default_branch(change_dir_base):
    """Test getting the default branch of current repo."""
    branch = get_default_branch('.')
    assert branch == 'main'


def test_get_latest_release(change_dir_base):
    """Get the latest release in repo."""
    release = get_latest_release('.')
    assert len(release) > 4



from tackle.main import tackle


def test_provider_tackle_run_hook(change_dir):
    """Check run_hook."""
    output = tackle()
    assert output['expanded']['foo'] == 'baz'
    assert output['compact']['foo'] == 'baz'



# Moved to utils as used in models.py


def wrap_braces_if_not_exist(value):
    """Wrap with braces if they don't exist."""
    if '{{' in value and '}}' in value:
        # Already templated
        return value
    return '{{' + value + '}}'


def wrap_jinja_braces(value):
    """Wrap a string with braces so it can be templated."""
    if isinstance(value, str):
        return wrap_braces_if_not_exist(value)
    # Nothing else can be wrapped
    return value



import os
import pytest
from tackle.main import tackle


@pytest.fixture()
def clean_outputs(change_dir):
    """Remove all the files prefixed with output before and after test."""
    files = [f for f in os.listdir() if f.split('.')[0].startswith('output')]
    for f in files:
        os.remove(f)
    yield
    files = [f for f in os.listdir() if f.split('.')[0].startswith('output')]
    for f in files:
        os.remove(f)


def test_provider_ini_read(change_dir, clean_outputs):
    output = tackle()
    assert os.path.exists('output.ini')
    assert output['read']['section']['stuff'] == 'things'
    assert output['read']['another']['foo'] is None



"""
Utils for handling command arguments along with unpacking hook arguments. Used in cli
to unpack input args/kwargs and in parser to unpack hook calls -
i.e. key->: hook arg --kwarg thing
"""
import ast
import re


def strip_dashes(raw_arg: str) -> str:
    """Remove dashes prepended to string for arg splitting."""
    while raw_arg.startswith('-'):
        raw_arg = raw_arg[1:]
    return raw_arg


def literal_eval(input):
    """
    Wrapper for ast.literal eval that accounts for hex and bools which it normally
     coerces to int/string respectively.
    """
    if isinstance(input, str) and input.startswith('0x'):
        # Prevent hex from being coerced to int type
        return input
    else:
        try:
            return ast.literal_eval(input)
        except (ValueError, SyntaxError):
            # Fix ast.literal_eval coercing bools to strings
            if input == 'true':
                return True
            elif input == 'false':
                return False
            else:
                return input


def split_input_string(input_string: str) -> list:
    """
    Split first on whitespace then regex each item to qualify if it needs to be
    interpreted as literal.
    """
    # Inspired from https://stackoverflow.com/a/524796/12642712
    # Split on whitespace or `=`
    # When quotes are preceded by `,:{[(=`, ignore
    # Don't split between quotes with `,:]})`
    # Otherwise split on quotes
    # Repeated for single quotes
    input_list = [
        i
        for i in re.split(
            "( |(?<!,|\:|\(|\{|\[|=)\\\"(?!\,|\:|\)|\}|\]).*?\\\"(?!\}|\])|(?<!\,|\:|\(|\{|\[|=)'(?!,|\:|\)|\}|\]).*?'(?!\}|\]))",  # noqa
            input_string,
        )
        if i.strip()
    ]

    # Remove the '=' that are split out into their own item
    input_list = [i for i in input_list if i != '=']

    output = []
    for i in input_list:
        output.append(literal_eval(i))
    return output


def unpack_args_kwargs_string(input_string: str) -> (list, dict, list):
    """Split up based on whitespace input args and pass to unpack_args_kwargs_list."""
    input_list = split_input_string(input_string)
    return unpack_args_kwargs_list(input_list)


def assert_if_flag(arg: str):
    FLAG_REGEX = re.compile(
        r"""^[\-|\-\-]+[a-zA-Z0-9]""",
        re.VERBOSE,
    )
    return bool(FLAG_REGEX.match(arg))


def unpack_args_kwargs_list(input_list: list) -> (list, dict, list):
    """
    Take the input template and unpack the args and kwargs if they exist.
    Updates the command_args and command_kwargs with a list of strings and
    list of dicts respectively.
    """
    input_list_length = len(input_list)
    args = []
    kwargs = {}
    flags = []

    i = 0
    while i < input_list_length:
        raw_arg = input_list[i]

        # Look ahead for if the last item is a flag
        if i + 1 < input_list_length:
            next_raw_arg = input_list[i + 1]
        else:
            # Hack for asserting if a flag - two items starting with a dash or last item
            next_raw_arg = "--hack"

        if isinstance(raw_arg, str):
            # if raw_arg.startswith('-'):
            if assert_if_flag(raw_arg):
                if isinstance(next_raw_arg, str):
                    # if next_raw_arg.startswith('-'):
                    if assert_if_flag(next_raw_arg):
                        # Field is a flag
                        flags.append(strip_dashes(raw_arg))
                    else:
                        # Field is a kwarg
                        kwargs.update({strip_dashes(raw_arg): input_list[i + 1]})
                        i += 1
                else:
                    # Field is a kwarg
                    kwargs.update({strip_dashes(raw_arg): input_list[i + 1]})
                    i += 1
            else:
                # Field is an argument
                args.append(raw_arg)
        else:
            # Field is an argument
            args.append(raw_arg)

        i += 1
    return args, kwargs, flags



"""Helper functions for working with version control systems."""
import logging
import os
import subprocess
from shutil import which

from tackle.exceptions import (
    RepositoryCloneFailed,
    RepositoryNotFound,
    VCSNotInstalled,
    VersionNotFoundError,
)
from tackle.utils.prompts import prompt_and_delete
from tackle.utils.paths import make_sure_path_exists
from tackle.settings import settings
from tackle.utils.paths import work_in

logger = logging.getLogger(__name__)

BRANCH_ERRORS = [
    'error: pathspec',
    'unknown revision',
]


# def identify_repo(repo_url):
#     """Determine if `repo_url` should be treated as a URL to a git or hg repo.
#
#     Repos can be identified by prepending "hg+" or "git+" to the repo URL.
#
#     :param repo_url: Repo URL of unknown type.
#     :returns: ('git', repo_url), ('hg', repo_url), or None.
#     """
#     repo_url_values = repo_url.split('+')
#     if len(repo_url_values) == 2:
#         repo_type = repo_url_values[0]
#         if repo_type in ["git", "hg"]:
#             return repo_type, repo_url_values[1]
#         else:
#             raise UnknownRepoType
#     else:
#         if 'git' in repo_url:
#             return 'git', repo_url
#         elif 'bitbucket' in repo_url:
#             return 'hg', repo_url
#         else:
#             raise UnknownRepoType


# def prompt_and_delete(path, no_input=False):
#     """
#     Ask user if it's okay to delete the previously-downloaded file/directory.
#
#     If yes, delete it. If no, checks to see if the old version should be
#     reused. If yes, it's reused; otherwise, Tackle exits.
#
#     :param path: Previously downloaded zipfile.
#     :param no_input: Suppress prompt to delete repo and just delete it.
#     :return: True if the content was deleted
#     """
#     # Suppress prompt if called via API
#     if no_input:
#         ok_to_delete = True
#     else:
#         question = (
#             "You've downloaded {} before. Is it okay to delete and re-download it?"
#         ).format(path)
#
#         ok_to_delete = read_user_yes_no(question, 'no')
#
#     if ok_to_delete:
#         if os.path.isdir(path):
#             rmtree(path)
#         else:
#             os.remove(path)
#         return True
#     else:
#         ok_to_reuse = read_user_yes_no(
#             "Do you want to re-use the existing version?", 'yes'
#         )
#
#         if ok_to_reuse:
#             return False
#
#         sys.exit()


def clone(repo_url, checkout=None, clone_to_dir='.', no_input=False):
    """Clone a repo to the current directory.

    :param repo_url: Repo URL of unknown type.
    :param checkout: The branch, tag or commit ID to checkout after clone.
    :param clone_to_dir: The directory to clone to.
                         Defaults to the current directory.
    :param no_input: Suppress all user prompts when calling via API.
    :returns: str with path to the new directory of the repository.
    """
    # Ensure that clone_to_dir exists
    clone_to_dir = os.path.expanduser(clone_to_dir)
    make_sure_path_exists(clone_to_dir)

    # identify the repo_type
    # repo_type, repo_url = identify_repo(repo_url)

    # check that the appropriate VCS for the repo_type is installed
    if not is_vcs_installed('git'):
        msg = "git is not installed."
        raise VCSNotInstalled(msg)

    repo_url = repo_url.rstrip('/')
    repo_name = os.path.split(repo_url)[1]
    repo_name = repo_name.split(':')[-1].rsplit('.git')[0]
    repo_dir = os.path.normpath(os.path.join(clone_to_dir, repo_name))
    logger.debug('repo_dir is {0}'.format(repo_dir))

    if os.path.isdir(repo_dir):
        clone = prompt_and_delete(repo_dir, no_input=no_input)
    else:
        clone = True

    if clone:
        try:
            subprocess.check_output(
                ['git', 'clone', repo_url],
                cwd=clone_to_dir,
                stderr=subprocess.STDOUT,
            )
            if checkout is not None:
                subprocess.check_output(
                    ['git', 'checkout', checkout],
                    cwd=repo_dir,
                    stderr=subprocess.STDOUT,
                )
        except subprocess.CalledProcessError as clone_error:
            output = clone_error.output.decode('utf-8')
            if 'not found' in output.lower():
                raise RepositoryNotFound(
                    'The repository {} could not be found, '
                    'have you made a typo?'.format(repo_url)
                )
            if any(error in output for error in BRANCH_ERRORS):
                raise RepositoryCloneFailed(
                    'The {} branch of repository {} could not found, '
                    'have you made a typo?'.format(checkout, repo_url)
                )
            raise

    return repo_dir


def is_vcs_installed(repo_type):
    """
    Check if the version control system for a repo type is installed.

    :param repo_type:
    """
    return bool(which(repo_type))


def get_default_branch(repo_path):
    """Get the default branch from a repo / provider."""
    cmd = "git remote show origin"
    with work_in(repo_path):
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            shell=True,
        )
        stdout, stderr = p.communicate()
        if p.returncode == 0:
            for i in stdout.strip().splitlines():
                if i.startswith(b'  HEAD'):
                    return i.split()[2].decode("utf-8")
        else:
            raise ValueError("No idea why this would not work....")


def get_latest_release(repo_path):
    """Checkout the latest release."""
    cmd = "git ls-remote --tags --exit-code --refs origin"
    with work_in(repo_path):
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            shell=True,
        )
        stdout, stderr = p.communicate()
        if p.returncode == 0:
            # last line is the latest release
            return stdout.strip().splitlines()[-1].decode('utf-8').split('/')[-1]
        elif p.returncode == 2:
            return None


def get_repo_version(repo_path):
    """Get the current repo version / branch."""
    with work_in(repo_path):
        with open('.git/HEAD') as f:
            ref = f.read()
    ref_list = ref.strip().split('/')
    if len(ref_list) == 1:
        return ref_list[0]
    return ref_list[2]


def checkout_version(repo_path, version):
    """Checkout a version / branch."""
    cmd = f"git checkout {version}"
    with work_in(repo_path):
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            shell=True,
        )
        stdout, stderr = p.communicate()

        if 'Your branch is behind' in str(stdout):
            cmd = 'git pull'
            p = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                shell=True,
            )
            p.communicate()

        if p.returncode == 0:
            return
            # return stdout.strip()
        else:
            rough_repo = '/'.join(repo_path.split('/')[-2:])
            raise VersionNotFoundError(
                f"Could not find the version='{version}' for the repo {rough_repo}."
            ) from None


def parse_repo_ref(repo):
    """Parse the repo string into parts."""
    # If starts with https
    organization = None
    provider_name = None
    repo_prefix = None

    if repo.startswith('https://'):
        repo_parts = repo.split('/')
        repo_prefix = 'https://' + repo_parts[2]
        organization = repo_parts[3]
        provider_name = repo_parts[4]

    elif len(repo.split('/')) == 2:
        # Handle type robcxyz/tackle-box
        repo_parts = repo.split('/')
        repo_prefix = 'https://github.com'
        organization = repo_parts[0]
        provider_name = repo_parts[1]

    elif repo.split('.')[1].startswith('com'):
        # Handles type github.com/robcxyz/tackle-box
        # TODO: This condition can be tightened
        repo_parts = repo.split('/')
        repo_prefix = 'https://' + repo_parts[0]
        organization = repo_parts[1]
        provider_name = repo_parts[2]

    if not organization or not provider_name:
        # This should be found by now otherwise it is an invalid format
        # Should not happen as the regex that qualifies a repo shouldn't have passed
        raise ValueError

    repo_url = '/'.join([repo_prefix, organization, provider_name])
    return repo_url, organization, provider_name


def get_repo_source(repo: str, repo_version: str = None) -> str:
    """Clone a provider into the providers dir with checkout out the right version."""
    repo_url, organization, provider_name = parse_repo_ref(repo)
    # provider_dir = os.path.join(settings.provider_dir, organization, provider_name)
    org_dir = os.path.join(settings.provider_dir, organization)
    provider_dir = os.path.join(org_dir, provider_name)

    logger.debug(f"Repo {repo_url} found for org {organization}, name {provider_name}.")

    if not os.path.exists(provider_dir):
        # Clone if dir does not exist (new provider)
        clone(repo_url, clone_to_dir=org_dir)
    else:
        subprocess.check_output(
            ['git', 'fetch'],
            cwd=provider_dir,
            stderr=subprocess.STDOUT,
        )

    # Get the default branch of the repo
    # If a version is specified, then check that first
    version = get_repo_version(provider_dir)
    logger.debug(f"Version for {provider_name} equal to {version}.")

    if repo_version:
        if repo_version == 'latest':
            default_branch = get_default_branch(repo_path=provider_dir)
            checkout_version(repo_path=provider_dir, version=default_branch)
        else:
            checkout_version(repo_path=provider_dir, version=repo_version)
    else:
        latest_release = get_latest_release(repo_path=provider_dir)
        if version != latest_release and latest_release is not None:
            checkout_version(repo_path=provider_dir, version=latest_release)

    return provider_dir



#!/usr/bin/env python
"""Tackle distutils configuration."""
import sys
import os
import re
from collections import defaultdict
import codecs

from setuptools import setup, find_packages


with open('README.md', encoding='utf-8') as readme_file:
    readme = readme_file.read()


def _read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    """Get the version from the __version__ file in the tackle dir."""
    for line in _read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def parse_requirements_file(reqs_file, provider_requirements, key):
    """
    Parse a requirements file and update the provider_requirements dict.

    :param reqs_file: Path to requirements file
    :param provider_requirements: A default dict with keys for `extras_require` and value
        as set of requirements.
    :param key: Key for installing `extras_require`.
    :return: Response of request in a dict

    """
    with open(reqs_file) as fp:
        for k in fp:
            if k.strip() and not k.startswith('#'):
                tags = set()
                if ':' in k:
                    k, v = k.split(':')
                    tags.update(vv.strip() for vv in v.split(','))
                tags.add(re.split('[<=>]', k)[0])

                for _ in tags:
                    provider_requirements[key].add(k)


def get_provider_requirements():
    """Get the provider requirements from each provider to allow extra requirements."""
    providers_dir = os.path.join(os.path.dirname(__file__), 'tackle', 'providers')
    providers = os.listdir(providers_dir)
    provider_requirements = defaultdict(set)
    for p in providers:
        if p == '__init__.py':
            continue
        reqs_file = os.path.join(providers_dir, p, 'requirements.txt')
        if os.path.isfile(reqs_file):
            parse_requirements_file(reqs_file, provider_requirements, p)
    parse_requirements_file('requirements-dev.txt', provider_requirements, 'dev')
    provider_requirements['all'] = set(
        vv for v in provider_requirements.values() for vv in v
    )
    return provider_requirements


INSTALL_REQUIREMENTS = [
    'Jinja2>3.0.0',
    # 'requests>=2.23.0',  # This would be needed if we allowed url sources
    'pydantic>=1.8.0',
    'InquirerPy>=0.3.3',
    'ruamel.yaml>=0.17.0',
    'rich>=12.6.0',
    'xdg==5.1.1',
]

if sys.argv[-1] == 'readme':
    print(readme)
    sys.exit()

setup(
    name='tackle-box',
    version=get_version(os.path.join('tackle', '__init__.py')),
    description=(
        'Tackle box is a declarative DSL for building modular workflows and code '
        'generators. Tool is plugins based and can easily be extended by writing '
        'additional hooks or importing external providers creating a web of '
        'interoperable CLIs.'
    ),
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Rob Cannon',
    author_email='robc.io.opensource@gmail.com',
    url='https://github.com/robcxyz/tackle-box',
    packages=find_packages(exclude=['tests*', 'logo*', 'docs*', '.github*']),
    package_dir={'tackle': 'tackle'},
    entry_points={'console_scripts': ['tackle = tackle.cli:main']},
    extras_require=get_provider_requirements(),
    include_package_data=True,
    python_requires='>=3.7',
    install_requires=INSTALL_REQUIREMENTS,
    license='BSD',
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python",
        "Topic :: Software Development",
    ],
    keywords=[
        "cookiecutter",
        "tackle",
        "tackle-box",
        "tacklebox",
        "tackle box",
        "Python",
        "projects",
        "project templates",
        "Jinja2",
        "skeleton",
        "scaffolding",
        "project directory",
        "package",
        "packaging",
        "kubernetes",
    ],
)
